"""
Test class level functionality.
"""

import pickle
import time
from multiprocessing.pool import ThreadPool
from typing import Dict, Union

import fsspec.core
import pytest

from ossfs import AioOSSFileSystem, OSSFileSystem
from ossfs.base import BaseOSSFileSystem
from tests.conftest import function_name


@pytest.mark.parametrize("aio", [False, True])
@pytest.mark.parametrize("default_cache_type", ["none", "bytes", "readahead"])
def test_default_cache_type(
    init_config: Dict, default_cache_type: str, test_path: str, aio: bool
):
    path = f"{test_path}/{function_name()}/"
    data = b"a" * (10 * 2**20)
    file = path + "/test_default_cache_type/file"
    init_config["default_cache_type"] = default_cache_type
    if aio:  # noqa: SIM108
        ossfs = AioOSSFileSystem(**init_config)
    else:
        ossfs = OSSFileSystem(**init_config)
    with ossfs.open(file, "wb") as f_wb:
        f_wb.write(data)

    with ossfs.open(file, "rb") as f_rb:
        assert isinstance(f_rb.cache, fsspec.core.caches[default_cache_type])
        out = f_rb.read(len(data))
        assert len(data) == len(out)
        assert out == data


@pytest.mark.parametrize("ossfs", ["sync", "async"], indirect=True)
@pytest.mark.parametrize("cache_type", ["none", "bytes", "readahead"])
def test_cache_type(
    ossfs: Union["OSSFileSystem", "AioOSSFileSystem"], cache_type: str, test_path: str
):
    path = f"{test_path}/{function_name(ossfs)}/"
    data = b"a" * (10 * 2**20)
    file = path + "/test_cache_type/file"

    with ossfs.open(file, "wb") as f_wb:
        f_wb.write(data)

    with ossfs.open(file, "rb", cache_type=cache_type) as f_rb:
        assert isinstance(f_rb.cache, fsspec.core.caches[cache_type])
        out = f_rb.read(len(data))
        assert len(data) == len(out)
        assert out == data


@pytest.mark.parametrize("ossfs", ["sync", "async"], indirect=True)
def test_current(ossfs: Union["OSSFileSystem", "AioOSSFileSystem"], init_config: Dict):
    ossfs._cache.clear()
    ossfs = OSSFileSystem(**init_config)
    assert ossfs.current() is ossfs
    assert OSSFileSystem.current() is ossfs


def test_connect_many(init_config: Dict, test_bucket_name: str):
    def task(num):
        ossfs = OSSFileSystem(**init_config)
        ossfs.ls(test_bucket_name)
        time.sleep(5)
        ossfs.ls(test_bucket_name)
        return True

    pool = ThreadPool(processes=20)
    out = pool.map(task, range(40))
    assert all(out)
    pool.close()
    pool.join()


@pytest.mark.parametrize("ossfs", ["sync", "async"], indirect=True)
def test_pickle(ossfs: Union["OSSFileSystem", "AioOSSFileSystem"], test_path: str):
    path = f"{test_path}/{function_name(ossfs)}/"
    for number in range(10):
        ossfs.touch(path + "file" + str(number))

    ossfs1 = pickle.loads(pickle.dumps(ossfs))
    assert ossfs.ls(path) == ossfs1.ls(path)
    ossfs2 = pickle.loads(pickle.dumps(ossfs1))
    assert ossfs.ls(path) == ossfs2.ls(path)


def test_strip_protocol():
    """
    Test protocols
    """
    address = "http://oss-cn-hangzhou.aliyuncs.com/mybucket/myobject"
    assert BaseOSSFileSystem._strip_protocol(address) == "/mybucket/myobject"
    address = "oss://mybucket/myobject"
    assert BaseOSSFileSystem._strip_protocol(address) == "/mybucket/myobject"



"""OSS error codes adapted into more natural Python ones.
Adapted from:
    https://error-center.alibabacloud.com/status/product/Oss?spm=a2c63.p38356.879954.5.3e172c31eo6sN9
"""

import errno

from oss2.exceptions import OssError

ERROR_CODE_TO_EXCEPTION = {
    "NoSuchBucket": FileNotFoundError,
    "NoSuchKey": FileNotFoundError,
    "NotFound": FileNotFoundError,
    "AccessDenied": PermissionError,
}


def translate_oss_error(
    error: OssError, *args, message=None, set_cause=True, **kwargs
) -> BaseException:
    """Convert a ClientError exception into a Python one.
    Parameters
    ----------
    error : oss2.exceptions.OssError
        The exception returned by the OSS Server.
    message : str
        An error message to use for the returned exception. If not given, the
        error message returned by the server is used instead.
    set_cause : bool
        Whether to set the __cause__ attribute to the previous exception if the
        exception is translated.
    *args, **kwargs :
        Additional arguments to pass to the exception constructor, after the
        error message. Useful for passing the filename arguments to
        ``IOError``.
    Returns
    -------
    An instantiated exception ready to be thrown. If the error code isn't
    recognized, an IOError with the original error message is returned.
    """
    if not isinstance(error, OssError):
        # not a oss error:
        return error
    code = error.__class__.__name__
    constructor = ERROR_CODE_TO_EXCEPTION.get(code)
    if constructor:
        if not message:
            message = error.message
        custom_exc = constructor(message, *args, **kwargs)
    else:
        # No match found, wrap this in an IOError with the appropriate message.
        custom_exc = OSError(errno.EIO, message or str(error), *args)

    if set_cause:
        custom_exc.__cause__ = error
    return custom_exc



"""
Pytest setup
"""

import inspect
import os
import pathlib
import subprocess
import uuid
from typing import Dict, Optional, Union

import oss2
import pytest
import requests
from oss2 import Auth

from ossfs import AioOSSFileSystem, OSSFileSystem

PORT = 5555
AccessKeyId = os.environ.get("OSS_ACCESS_KEY_ID", "")
AccessKeySecret = os.environ.get("OSS_SECRET_ACCESS_KEY", "")
LICENSE_PATH = os.path.join(pathlib.Path(__file__).parent.parent.resolve(), "LICENSE")
NUMBERS = b"1234567890\n"


test_id = uuid.uuid4()


def bucket_relative_path(path: str) -> str:
    return path.split("/", 2)[2]


def function_name(ossfs: Optional[Union["OSSFileSystem", "AioOSSFileSystem"]] = None):
    file_name = inspect.stack()[1].filename.rsplit(os.sep, maxsplit=1)[-1][:-3]
    function_name = inspect.stack()[1][0].f_code.co_name
    if ossfs:
        function_name += f"_{ossfs.__class__.__name__}"
    return f"{file_name}/{function_name}"


@pytest.fixture(scope="session")
def emulator_endpoint() -> str:
    return f"http://127.0.0.1:{PORT}/"


@pytest.fixture(scope="session")
def endpoint() -> str:
    endpoint = os.environ.get("OSS_ENDPOINT")
    assert endpoint
    return endpoint


@pytest.fixture(scope="session")
def test_bucket_name() -> str:
    test_bucket_name = os.environ.get("OSS_TEST_BUCKET_NAME")
    assert test_bucket_name
    return test_bucket_name


@pytest.fixture(scope="session", name="test_path")
def file_level_path(test_bucket_name: str, test_directory: str) -> str:
    return f"/{test_bucket_name}/{test_directory}"


@pytest.fixture(scope="session")
def anonymous_bucket_name() -> str:
    anonymous_bucket_name = os.environ.get("OSS_TEST_ANONYMOUS_BUCKET_NAME")
    assert anonymous_bucket_name
    return anonymous_bucket_name


@pytest.fixture(scope="session")
def test_directory() -> str:
    return f"ossfs_test/{test_id}"


@pytest.fixture
def oss_emulator_server_start(emulator_endpoint):
    """
    Start a local emulator server
    """
    with subprocess.Popen(f"ruby bin/emulator -r store -p {PORT}"):
        try:
            result = requests.get(emulator_endpoint, timeout=5)
            if result.ok:
                yield
        except TimeoutError as err:
            raise Exception("emulator start timeout") from err


@pytest.fixture(scope="session")
def init_config(endpoint) -> Dict:
    result = {}
    result["key"] = AccessKeyId
    result["secret"] = AccessKeySecret
    result["endpoint"] = endpoint
    return result


@pytest.fixture
def ossfs(request, init_config: Dict) -> Union["OSSFileSystem", "AioOSSFileSystem"]:
    if hasattr(request, "param") and request.param == "async":
        ossfs = AioOSSFileSystem(**init_config)
    else:
        ossfs = OSSFileSystem(**init_config)
    ossfs.invalidate_cache()
    return ossfs


@pytest.fixture(scope="session")
def auth():
    return Auth(AccessKeyId, AccessKeySecret)


@pytest.fixture(scope="session")
def bucket(auth: "Auth", endpoint: str, test_bucket_name: str):
    return oss2.Bucket(auth, endpoint, test_bucket_name)


@pytest.fixture(scope="session")
def file_in_anonymous(auth: "Auth", endpoint: str, anonymous_bucket_name: str) -> str:
    file = "file"
    bucket = oss2.Bucket(auth, endpoint, anonymous_bucket_name)
    bucket.put_object(file, "foobar")
    return f"/{anonymous_bucket_name}/{file}"


@pytest.fixture(scope="session")
def number_file(
    bucket: "oss2.Bucket", test_directory: str, test_bucket_name: str
) -> str:
    filename = f"{test_directory}/number"
    bucket.put_object(filename, NUMBERS)
    return f"/{test_bucket_name}/{filename}"


@pytest.fixture(scope="session")
def license_file(
    bucket: "oss2.Bucket", test_bucket_name: str, test_directory: str
) -> str:
    filename = f"{test_directory}/LICENSE"
    bucket.put_object_from_file(filename, LICENSE_PATH)
    return f"/{test_bucket_name}/{filename}"



"""
OSSFS
----------------------------------------------------------------
A pythonic file-systems interface to OSS (Object Storage Service)
"""

from .async_oss import AioOSSFileSystem
from .core import OSSFileSystem
from .file import OSSFile

__all__ = ["AioOSSFileSystem", "OSSFile", "OSSFileSystem"]



"""
Code of AioOSSFileSystem
"""

import inspect
import logging
import os
import weakref
from datetime import datetime
from hashlib import sha256
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

import aiooss2
from aiooss2 import AioBucket, AioService, AnonymousAuth
from aiooss2.http import AioSession
from fsspec.asyn import AsyncFileSystem, _run_coros_in_chunks, sync, sync_wrapper
from fsspec.exceptions import FSTimeoutError
from oss2.exceptions import ClientError, OssError, RequestError
from oss2.models import PartInfo

from .base import (
    DEFAULT_BLOCK_SIZE,
    DEFAULT_POOL_SIZE,
    SIMPLE_TRANSFER_THRESHOLD,
    BaseOSSFileSystem,
)
from .exceptions import translate_oss_error
from .utils import as_progress_handler, async_prettify_info_result

if TYPE_CHECKING:
    from aiooss2.models import AioGetObjectResult
    from oss2.models import (
        AppendObjectResult,
        HeadObjectResult,
        InitMultipartUploadResult,
        ListBucketsResult,
        PutObjectResult,
        SimplifiedBucketInfo,
        SimplifiedObjectInfo,
    )

logger = logging.getLogger("ossfs")


class AioOSSFileSystem(BaseOSSFileSystem, AsyncFileSystem):
    """
    A pythonic file-systems interface to OSS (Object Storage Service)
    Base on async operations.

    Examples
    --------
    >>> ossfs = AioOSSFileSystem(anon=False)
    >>> ossfs.ls('my-bucket/')
    ['my-file.txt']

    >>> with ossfs.open('my-bucket/my-file.txt', mode='rb') as f:
    ...     print(f.read())
    b'Hello, world!'
    """

    protocol = "oss"

    def __init__(
        self,
        psize: int = DEFAULT_POOL_SIZE,
        **kwargs,
    ):
        """
        ----------------------------------------------------------------
        Addition arguments

        Args:
            psize (int, optional): concurrency number of the connections to
            the server. Defaults to DEFAULT_POOL_SIZE.
        """
        super().__init__(**kwargs)
        self._psize = psize
        self._session: Optional[AioSession] = None

    def _get_bucket(
        self, bucket_name: str, connect_timeout: Optional[int] = None
    ) -> AioBucket:
        """
        get the new aio bucket instance
        """
        if self._endpoint is None:
            raise ValueError("endpoint is required")
        try:
            return AioBucket(
                auth=self._auth,
                endpoint=self._endpoint,
                bucket_name=bucket_name,
                connect_timeout=connect_timeout,
                session=self._session,
                app_name="ossfs",
            )
        except ClientError as err:
            raise FileNotFoundError(bucket_name) from err

    async def set_session(self, refresh: bool = False):
        """Establish a connection session object.
        Returns
        -------
        Session to be closed later with await .close()
        """
        logger.debug("Connect AioSession instance")
        if self._session is None or self._session.closed or refresh:
            if self._session is None:
                self._session = AioSession(self._psize)
            await self._session.__aenter__()
            # the following actually closes the aiohttp connection; use of privates
            # might break in the future, would cause exception at gc time
            if not self.asynchronous:
                weakref.finalize(self, self.close_session)

    def close_session(self):
        """Close a connection session object."""
        if self._session is None or self._session.closed:
            return
        if self.loop is not None and self.loop.is_running():
            try:
                sync(self.loop, self._session.close, timeout=0.1)
                return
            except FSTimeoutError:
                pass

    async def _call_oss(
        self,
        method_name: str,
        *args,
        bucket: Optional[str] = None,
        timeout: Optional[int] = None,
        **kwargs,
    ):
        if self._endpoint is None:
            raise ValueError("endpoint is required")
        await self.set_session()
        if bucket:
            service: Union[AioService, AioBucket] = self._get_bucket(bucket, timeout)
        else:
            service = AioService(
                auth=self._auth,
                endpoint=self._endpoint,
                session=self._session,
                connect_timeout=timeout,
                app_name="ossfs",
            )
        method = getattr(service, method_name, None)
        try:
            if not method:
                method = getattr(aiooss2, method_name)
                logger.debug("CALL: %s - %s - %s", method.__name__, args, kwargs)
                if inspect.iscoroutinefunction(method):
                    out = await method(service, *args, **kwargs)
                else:
                    out = method(service, *args, **kwargs)
            else:
                logger.debug("CALL: %s - %s - %s", method.__name__, args, kwargs)
                out = await method(*args, **kwargs)
            return out
        except (RequestError, OssError) as err:
            error = err
        raise translate_oss_error(error) from error

    async def _ls_dir(
        self,
        path: str,
        refresh: bool = False,
        max_items: int = 100,
        delimiter: str = "/",
        prefix: str = "",
    ):
        norm_path = path.strip("/")
        if norm_path in self.dircache and not refresh and not prefix and delimiter:
            return self.dircache[norm_path]
        logger.debug("Get directory listing for %s", norm_path)
        bucket, key = self.split_path(norm_path)
        prefix = prefix or ""
        if key:
            prefix = f"{key}/{prefix}"

        files = [
            obj_dict
            async for obj_dict in self._iterdir(
                bucket,
                max_keys=max_items,
                delimiter=delimiter,
                prefix=prefix,
            )
        ]

        if not prefix and delimiter == "/":
            self.dircache[norm_path] = files
        return files

    async def _iterdir(
        self,
        bucket: str,
        max_keys: int = 100,
        delimiter: str = "/",
        prefix: str = "",
    ):
        """Iterate asynchronously over files and directories under `prefix`.

        The contents are yielded in arbitrary order as info dicts.
        """
        response = await self._call_oss(
            "AioObjectIterator",
            bucket=bucket,
            prefix=prefix,
            delimiter=delimiter,
            max_keys=max_keys,
        )
        obj: SimplifiedObjectInfo
        async for obj in response:
            data = self._transfer_object_info_to_dict(bucket, obj)
            yield data

    async def _ls_buckets(self, refresh: bool = False) -> List[Dict[str, Any]]:
        if "" not in self.dircache or refresh:
            if isinstance(self._auth, AnonymousAuth):
                logging.warning("cannot list buckets if not logged in")
                return []
            results: List[Dict[str, Any]] = []
            try:
                files: ListBucketsResult = await self._call_oss("list_buckets")
            except ClientError:
                # listbucket permission missing
                return []
            file: SimplifiedBucketInfo
            for file in files.buckets:
                data: Dict[str, Any] = {}
                data["name"] = file.name
                data["size"] = 0
                data["type"] = "directory"
                results.append(data)
            self.dircache[""] = results
        else:
            results = self.dircache[""]
        return results

    @async_prettify_info_result
    async def _ls(self, path: str, detail: bool = True, **kwargs):  # noqa: ARG002
        """List files in given bucket, or list of buckets.

        Listing is cached unless `refresh=True`.

        Note: only your buckets associated with the login will be listed by
        `ls('')`, not any public buckets (even if already accessed).

        Parameters
        ----------
        path : string/bytes
            location at which to list files
        refresh : bool (=False)
            if False, look in local cache for file details first
        """
        refresh = kwargs.pop("refresh", False)
        norm_path = self._strip_protocol(path).strip("/")
        if norm_path != "":
            files = await self._ls_dir(path, refresh)
            if not files and "/" in norm_path:
                files = await self._ls_dir(self._parent(path), refresh=refresh)
                files = [
                    file
                    for file in files
                    if file["name"].strip("/") == norm_path
                    and file["type"] != "directory"
                ]
        else:
            files = await self._ls_buckets(refresh)
        return files

    @async_prettify_info_result
    async def _info(self, path: str, **kwargs):  # noqa: C901, PLR0912
        norm_path = self._strip_protocol(path).lstrip("/")
        if norm_path == "":
            return {"name": path, "size": 0, "type": "directory"}
        bucket, key = self.split_path(norm_path)
        self._get_bucket(bucket)
        refresh = kwargs.pop("refresh", False)
        if not refresh:
            out = self._ls_from_cache(norm_path)
            if out is not None:
                out = [o for o in out if o["name"].strip("/") == norm_path]
                if out:
                    result = out[0]
                else:
                    result = {"name": norm_path, "size": 0, "type": "directory"}
                return result

        if key:
            try:
                obj_out: HeadObjectResult = await self._call_oss(
                    "head_object",
                    key,
                    bucket=bucket,
                )
                return {
                    "LastModified": obj_out.last_modified,
                    "size": obj_out.content_length,
                    "name": path,
                    "type": "file",
                }
            except (PermissionError, FileNotFoundError):
                pass
            # We check to see if the path is a directory by attempting to list its
            # contexts. If anything is found, it is indeed a directory
            try:
                ls_out = await self._call_oss(
                    "AioObjectIterator",
                    bucket=bucket,
                    prefix=key.rstrip("/") + "/",
                    delimiter="/",
                    max_keys=100,
                )
                try:
                    async for _ in ls_out:
                        return {
                            "size": 0,
                            "name": path,
                            "type": "directory",
                        }
                except OssError as err:
                    raise translate_oss_error(err) from err
            except (PermissionError, FileNotFoundError):
                pass

        else:
            for bucket_info in await self._ls_buckets():
                if bucket_info["name"] == norm_path.rstrip("/"):
                    return {
                        "size": 0,
                        "name": path,
                        "type": "directory",
                    }
        raise FileNotFoundError(path)

    def _cache_result_analysis(self, norm_path: str, parent: str) -> bool:
        if norm_path in self.dircache:
            for file_info in self.dircache[norm_path]:
                # For files the dircache can contain itself.
                # If it contains anything other than itself it is a directory.
                if file_info["name"] != norm_path:
                    return True
            return False

        for file_info in self.dircache[parent]:
            if file_info["name"] == norm_path:
                # If we find ourselves return whether we are a directory
                return file_info["type"] == "directory"
        return False

    async def _isdir(self, path: str) -> bool:
        norm_path = self._strip_protocol(path).strip("/")
        # Send buckets to super
        if norm_path == "":
            return True
        if "/" not in norm_path:
            for bucket_info in await self._ls_buckets():
                if bucket_info["name"] == norm_path:
                    return True
            return False

        parent = self._parent(norm_path)
        if norm_path in self.dircache or parent in self.dircache:
            return self._cache_result_analysis(norm_path, parent)

        # This only returns things within the path and NOT the path object itself
        try:
            return bool(await self._ls_dir(norm_path))
        except FileNotFoundError:
            return False

    async def _put_file(self, lpath: str, rpath: str, **kwargs):
        bucket, key = self.split_path(rpath)
        if os.path.isdir(lpath):
            if key:
                # don't make remote "directory"
                return
            await self._mkdir(lpath)
        else:
            callback = as_progress_handler(kwargs.pop("callback", None))
            if os.path.getsize(lpath) >= SIMPLE_TRANSFER_THRESHOLD:
                await self._call_oss(
                    "resumable_upload",
                    bucket=bucket,
                    key=key,
                    filename=lpath,
                    progress_callback=callback,
                )
            else:
                await self._call_oss(
                    "put_object_from_file",
                    bucket=bucket,
                    key=key,
                    filename=lpath,
                    progress_callback=callback,
                )

        self.invalidate_cache(self._parent(rpath))

    async def _get_file(self, rpath: str, lpath: str, **kwargs):
        """
        Copy single remote file to local
        """
        bucket, key = self.split_path(rpath)
        if await self._isdir(rpath):
            # don't make local "directory"
            return
        callback = as_progress_handler(kwargs.pop("callback", None))
        if await self._size(rpath) >= SIMPLE_TRANSFER_THRESHOLD:
            await self._call_oss(
                "resumable_download",
                bucket=bucket,
                key=key,
                filename=lpath,
                progress_callback=callback,
            )
        else:
            await self._call_oss(
                "get_object_to_file",
                bucket=bucket,
                key=key,
                filename=lpath,
                progress_callback=callback,
                **kwargs,
            )

    @async_prettify_info_result
    async def _find(
        self,
        path: str,
        maxdepth: Optional[int] = None,
        withdirs: bool = False,
        detail: bool = False,  # noqa: ARG002
        **kwargs,
    ):
        """List all files below path.
        Like posix ``find`` command without conditions

        Parameters
        ----------
        path : str
        maxdepth: int or None
            If not None, the maximum number of levels to descend
        withdirs: bool
            Whether to include directory paths in the output. This is True
            when used by glob, but users usually only want files.
        prefix: str
            Only return files that match ``^{path}/{prefix}`` (if there is an
            exact match ``filename == {path}/{prefix}``, it also will be included)
        """
        out = {}
        prefix = kwargs.pop("prefix", "")
        path = self._verify_find_arguments(path, maxdepth, withdirs, prefix)
        if prefix:
            for info in await self._ls_dir(path, delimiter="", prefix=prefix):
                out.update({info["name"]: info})
        else:
            async for _, dirs, files in self._walk(path, maxdepth, detail=True):
                if withdirs:
                    files.update(dirs)
                out.update({info["name"]: info for _, info in files.items()})
            if await self._isfile(path) and path not in out:
                # walk works on directories, but find should also return [path]
                # when path happens to be a file
                out[path] = {}
        names = sorted(out)
        return {name: out[name] for name in names}

    async def _bulk_delete(self, pathlist, **kwargs):
        """
        Remove multiple keys with one call

        Parameters
        ----------
        pathlist : list(str)
            The keys to remove, must all be in the same bucket.
            Must have 0 < len <= 1000
        """
        if not pathlist:
            return
        bucket, key_list = self._get_batch_delete_key_list(pathlist)
        await self._call_oss("batch_delete_objects", key_list, bucket=bucket)

    async def _rm_file(self, path: str, **kwargs):
        bucket, key = self.split_path(path)
        await self._call_oss("delete_object", bucket=bucket, key=key)
        self.invalidate_cache(self._parent(path))

    async def _rm(self, path, recursive=False, batch_size=1000, **kwargs):
        if isinstance(path, list):
            for file in path:
                await self._rm(file)
            return

        paths = await self._expand_path(path, recursive=recursive)
        await _run_coros_in_chunks(
            [
                self._bulk_delete(paths[i : i + batch_size])
                for i in range(0, len(paths), batch_size)
            ],
            batch_size=3,
            nofiles=True,
        )

    async def _checksum(self, path, refresh=True):
        """
        Unique value for current version of file

        If the checksum is the same from one moment to another, the contents
        are guaranteed to be the same. If the checksum changes, the contents
        *might* have changed.

        Parameters
        ----------
        path : string/bytes
            path of file to get checksum for
        refresh : bool (=False)
            if False, look in local cache for file details first

        """
        return sha256(
            (
                str(await self._ukey(path))
                + str(await self._info(path, refresh=refresh))
            ).encode()
        ).hexdigest()

    checksum = sync_wrapper(_checksum)

    async def _ukey(self, path: str):
        """Hash of file properties, to tell if it has changed"""
        bucket_name, obj_name = self.split_path(path)
        obj_stream = await self._call_oss("get_object", obj_name, bucket=bucket_name)
        return obj_stream.server_crc

    checksum = sync_wrapper(_checksum)  # noqa: PIE794

    async def _cp_file(self, path1: str, path2: str, **kwargs):
        """Copy file between locations on OSS.

        preserve_etag: bool
            Whether to preserve etag while copying. If the file is uploaded
            as a single part, then it will be always equalivent to the md5
            hash of the file hence etag will always be preserved. But if the
            file is uploaded in multi parts, then this option will try to
            reproduce the same multipart upload while copying and preserve
            the generated etag.
        """
        bucket2, key2 = self.split_path(path2)
        bucket1, key1 = self.split_path(path1)
        self.invalidate_cache(self._parent(path2))
        if bucket1 != bucket2:
            tempdir = "." + self.ukey(path1)
            await self._get_file(path1, tempdir, **kwargs)
            await self._put_file(tempdir, path2, **kwargs)
            os.remove(tempdir)
        else:
            connect_timeout = kwargs.pop("connect_timeout", None)
            await self._call_oss(
                "copy_object",
                bucket1,
                key1,
                key2,
                bucket=bucket1,
                timeout=connect_timeout,
            )

    async def _append_object(self, path: str, location: int, value: bytes) -> int:
        """
        Append bytes to the object
        """
        bucket, key = self.split_path(path)
        result: AppendObjectResult = await self._call_oss(
            "append_object",
            key,
            location,
            value,
            bucket=bucket,
        )
        return result.next_position

    append_object = sync_wrapper(_append_object)

    async def _get_object(self, path: str, start: int, end: int) -> bytes:
        """
        Return object bytes in range
        """
        headers = {"x-oss-range-behavior": "standard"}
        bucket, key = self.split_path(path)
        object_stream: AioGetObjectResult = await self._call_oss(
            "get_object",
            key,
            bucket=bucket,
            byte_range=(start, end),
            headers=headers,
        )
        results = b""
        while True:
            result = await object_stream.read()
            if result:
                results += result
            else:
                break
        return results

    get_object = sync_wrapper(_get_object)

    async def _pipe_file(self, path: str, value: Union[str, bytes], **kwargs):
        bucket, key = self.split_path(path)
        self.invalidate_cache(path)
        block_size = kwargs.get("block_size", DEFAULT_BLOCK_SIZE)
        # 5 GB is the limit for an OSS PUT
        if len(value) < min(5 * 2**30, 2 * block_size):
            await self._call_oss("put_object", key, value, bucket=bucket, **kwargs)
            return
        init_multi_part_upload_result: InitMultipartUploadResult = await self._call_oss(
            "init_multipart_upload", key, bucket=bucket, **kwargs
        )
        parts: List[PartInfo] = []
        for i, off in enumerate(range(0, len(value), block_size)):
            part_number = i + 1
            value_block = value[off : off + block_size]
            put_object_result: PutObjectResult = await self._call_oss(
                "upload_part",
                key,
                init_multi_part_upload_result.upload_id,
                part_number,
                value_block,
                bucket=bucket,
            )
            parts.append(
                PartInfo(
                    part_number,
                    put_object_result.etag,
                    size=len(value_block),
                    part_crc=put_object_result.crc,
                )
            )
        await self._call_oss(
            "complete_multipart_upload",
            key,
            init_multi_part_upload_result.upload_id,
            parts,
            bucket=bucket,
        )

    async def _cat_file(self, path: str, start=None, end=None, **kwargs):
        bucket, key = self.split_path(path)
        object_stream: AioGetObjectResult = await self._call_oss(
            "get_object",
            bucket=bucket,
            key=key,
            byte_range=(start, end),
            **kwargs,
        )

        results = b""
        while True:
            result = await object_stream.read()
            if not result:
                break
            results += result
        return results

    async def _modified(self, path: str):
        """Return the modified timestamp of a file as a datetime.datetime"""
        bucket_name, obj_name = self.split_path(path)
        if not obj_name or await self._isdir(path):
            raise NotImplementedError("bucket has no modified timestamp")
        object_meta = await self._call_oss(
            "get_object_meta", obj_name, bucket=bucket_name
        )
        return int(
            datetime.strptime(  # noqa: DTZ007
                object_meta.headers["Last-Modified"],
                "%a, %d %b %Y %H:%M:%S %Z",
            ).timestamp()
        )

    modified = sync_wrapper(_modified)



"""
Code of OSSFileSystem and OSSFile
"""

import logging
from typing import TYPE_CHECKING, Union

from fsspec.spec import AbstractBufferedFile

if TYPE_CHECKING:
    from .async_oss import AioOSSFileSystem
    from .core import OSSFileSystem


logger = logging.getLogger("ossfs")
logging.getLogger("oss2").setLevel(logging.CRITICAL)
logging.getLogger("urllib3").setLevel(logging.WARNING)


class OSSFile(AbstractBufferedFile):
    """A file living in OSSFileSystem"""

    fs: Union["OSSFileSystem", "AioOSSFileSystem"]
    loc: int

    def _upload_chunk(self, final: bool = False) -> bool:  # noqa: ARG002
        """Write one part of a multi-block file upload
        Parameters
        ==========
        final: bool
            This is the last block, so should complete file, if
            self.autocommit is True.
        """
        self.loc = self.fs.append_object(self.path, self.loc, self.buffer.getvalue())
        return True

    def _initiate_upload(self):
        """Create remote file/upload"""
        if "a" in self.mode:
            self.loc = 0
            if self.fs.exists(self.path):
                self.loc = self.fs.info(self.path)["size"]
        elif "w" in self.mode:
            # create empty file to append to
            self.loc = 0
            if self.fs.exists(self.path):
                self.fs.rm_file(self.path)

    def _fetch_range(self, start: int, end: int) -> bytes:
        """
        Get the specified set of bytes from remote
        Parameters
        ==========
        start: int
        end: int
        """
        start = max(start, 0)
        end = min(self.size, end)
        if start >= end or start >= self.size:
            return b""
        return self.fs.get_object(self.path, start, end)



"""
Test all OSSFile related methods
"""

import io
import os
from typing import TYPE_CHECKING, Union

import pytest

from tests.conftest import LICENSE_PATH, NUMBERS, bucket_relative_path, function_name

if TYPE_CHECKING:
    from oss2 import Bucket

    from ossfs import AioOSSFileSystem, OSSFileSystem


@pytest.mark.parametrize("ossfs", ["async", "sync"], indirect=True)
def test_simple_read(
    ossfs: Union["OSSFileSystem", "AioOSSFileSystem"], test_path: str, bucket: "Bucket"
):
    object_name = f"{test_path}/{function_name(ossfs)}"
    data = os.urandom(10 * 2**20)
    bucket.put_object(bucket_relative_path(object_name), data)

    with ossfs.open(object_name, "rb") as f_rb:
        out = f_rb.read(len(data))
    assert len(data) == len(out)
    assert out == data


@pytest.mark.parametrize("ossfs", ["async", "sync"], indirect=True)
def test_simple_write(
    ossfs: Union["OSSFileSystem", "AioOSSFileSystem"], test_path: str, bucket: "Bucket"
):
    object_name = f"{test_path}/{function_name(ossfs)}"
    data = os.urandom(10 * 2**20)

    with ossfs.open(object_name, "wb") as f_wb:
        f_wb.write(data)

    out = bucket.get_object(bucket_relative_path(object_name)).read()
    assert out == data


@pytest.mark.parametrize("ossfs", ["async", "sync"], indirect=True)
def test_seek(
    ossfs: Union["OSSFileSystem", "AioOSSFileSystem"], test_path: str, bucket: "Bucket"
):
    object_name = f"{test_path}/{function_name(ossfs)}"
    bucket.put_object(bucket_relative_path(object_name), b"123")

    with ossfs.open(object_name) as f_seek:
        f_seek.seek(1000)
        with pytest.raises(ValueError):
            f_seek.seek(-1)
        with pytest.raises(ValueError):
            f_seek.seek(-5, 2)
        with pytest.raises(ValueError):
            f_seek.seek(0, 10)
        f_seek.seek(0)
        assert f_seek.read(1) == b"1"
        f_seek.seek(0)
        assert f_seek.read(1) == b"1"
        f_seek.seek(3)
        assert f_seek.read(1) == b""
        f_seek.seek(-1, 2)
        assert f_seek.read(1) == b"3"
        f_seek.seek(-1, 1)
        f_seek.seek(-1, 1)
        assert f_seek.read(1) == b"2"
        for i in range(4):
            assert f_seek.seek(i) == i


@pytest.mark.parametrize("ossfs", ["async", "sync"], indirect=True)
def test_read_small(
    ossfs: Union["OSSFileSystem", "AioOSSFileSystem"],
    number_file: str,
    bucket: "Bucket",
):
    with ossfs.open(number_file, "rb", block_size=3) as f_rb:
        out = []
        while True:
            data = f_rb.read(2)
            if data == b"":
                break
            out.append(data)
        assert bucket.get_object(bucket_relative_path(number_file)).read() == b"".join(
            out
        )


@pytest.mark.parametrize("ossfs", ["async", "sync"], indirect=True)
def test_read_ossfs_block(
    ossfs: Union["OSSFileSystem", "AioOSSFileSystem"],
    license_file: str,
    number_file: str,
):
    with open(LICENSE_PATH, "rb") as f_r:
        data = f_r.read()
    lines = io.BytesIO(data).readlines()
    assert ossfs.read_block(license_file, 0, 10, b"\n") == lines[0]
    assert ossfs.read_block(license_file, 40, 10, b"\n") == lines[1]
    assert ossfs.read_block(license_file, 0, 80, b"\n") == lines[0] + lines[1]
    assert ossfs.read_block(license_file, 0, 120, b"\n") == b"".join(
        [lines[0], lines[1], lines[2]]
    )

    lines = io.BytesIO(NUMBERS).readlines()
    assert len(ossfs.read_block(number_file, 0, 5)) == 5
    assert len(ossfs.read_block(number_file, 4, 150)) == len(NUMBERS) - 4
    assert ossfs.read_block(number_file, 20, 25) == b""

    assert ossfs.read_block(number_file, 5, None) == ossfs.read_block(
        number_file, 5, 25
    )


@pytest.mark.parametrize("ossfs", ["async", "sync"], indirect=True)
@pytest.mark.parametrize("size", [2**10, 2**20, 10 * 2**20])
def test_write(
    ossfs: Union["OSSFileSystem", "AioOSSFileSystem"],
    test_path: str,
    size: int,
    bucket: "Bucket",
):
    object_name = f"{test_path}/{function_name(ossfs)}"
    data = os.urandom(size)
    with ossfs.open(object_name, "wb") as f_wb:
        f_wb.write(data)
    out = bucket.get_object(bucket_relative_path(object_name)).read()
    assert data == out
    assert ossfs.info(object_name)["Size"] == len(data)
    ossfs.open(object_name, "wb").close()
    assert ossfs.info(object_name)["Size"] == 0


@pytest.mark.parametrize("ossfs", ["async", "sync"], indirect=True)
def test_write_fails(ossfs: Union["OSSFileSystem", "AioOSSFileSystem"], test_path: str):
    object_name = f"{test_path}/{function_name(ossfs)}"
    ossfs.touch(object_name)
    with pytest.raises(ValueError):
        ossfs.open(object_name, "rb").write(b"hello")
    f_wb = ossfs.open(object_name, "wb")
    f_wb.close()
    with pytest.raises(ValueError):
        f_wb.write(b"hello")
    with pytest.raises(FileNotFoundError):
        ossfs.open("nonexistentbucket/temp", "wb").close()


@pytest.mark.parametrize("ossfs", ["async", "sync"], indirect=True)
def test_write_blocks(
    ossfs: Union["OSSFileSystem", "AioOSSFileSystem"], test_path: str, bucket: "Bucket"
):
    object_name = f"{test_path}/{function_name(ossfs)}"
    with ossfs.open(object_name, "wb") as f_wb:
        f_wb.write(os.urandom(2 * 2**20))
        assert f_wb.buffer.tell() == 2 * 2**20
        f_wb.flush()
        assert f_wb.buffer.tell() == 2 * 2**20
        f_wb.write(os.urandom(2 * 2**20))
        f_wb.write(os.urandom(2 * 2**20))
    out = bucket.get_object(bucket_relative_path(object_name)).read()
    assert len(out) == 6 * 2**20
    with ossfs.open(object_name, "wb", block_size=10 * 2**20) as f_wb:
        f_wb.write(os.urandom(15 * 2**20))
        assert f_wb.buffer.tell() == 0
    out = bucket.get_object(bucket_relative_path(object_name)).read()
    assert len(out) == 15 * 2**20


@pytest.mark.parametrize("ossfs", ["async", "sync"], indirect=True)
def test_readline(
    ossfs: Union["OSSFileSystem", "AioOSSFileSystem"],
    number_file: str,
    license_file: str,
):
    with ossfs.open(f"{number_file}", "rb") as f_rb:
        result = f_rb.readline()
        expected = NUMBERS
        assert result == expected

    with ossfs.open(f"{license_file}", "rb") as f_rb, open(LICENSE_PATH, "rb") as f_l:
        result = f_rb.readline()
        expected = f_l.readline()
        assert result == expected


@pytest.mark.parametrize("ossfs", ["async", "sync"], indirect=True)
def test_readline_empty(
    ossfs: Union["OSSFileSystem", "AioOSSFileSystem"], test_path: str
):
    file = f"{test_path}/{function_name(ossfs)}/empty"
    data = b""
    with ossfs.open(file, "wb") as f_wb:
        f_wb.write(data)
    with ossfs.open(file, "rb") as f_rb:
        result = f_rb.readline()
        assert result == data


@pytest.mark.parametrize("ossfs", ["async", "sync"], indirect=True)
def test_readline_blocksize(
    ossfs: Union["OSSFileSystem", "AioOSSFileSystem"], test_path: str
):
    test_file_a = f"{test_path}/{function_name(ossfs)}/a"
    data = b"ab\n" + b"a" * (10 * 2**20) + b"\nab"
    with ossfs.open(test_file_a, "wb") as f_wb:
        f_wb.write(data)
    with ossfs.open(test_file_a, "rb") as f_rb:
        result = f_rb.readline()
        expected = b"ab\n"
        assert result == expected

        result = f_rb.readline()
        expected = b"a" * (10 * 2**20) + b"\n"
        assert result == expected

        result = f_rb.readline()
        expected = b"ab"
        assert result == expected


@pytest.mark.parametrize("ossfs", ["async", "sync"], indirect=True)
def test_next(ossfs: Union["OSSFileSystem", "AioOSSFileSystem"], license_file: str):
    with open(LICENSE_PATH, "rb") as f_rb, ossfs.open(license_file) as f_r:
        expected = f_rb.readline()
        result = next(f_r)
        assert result == expected


@pytest.mark.parametrize("ossfs", ["async", "sync"], indirect=True)
def test_iterable(ossfs: Union["OSSFileSystem", "AioOSSFileSystem"], test_path: str):
    file = f"{test_path}/{function_name(ossfs)}/file"
    data = b"abc\n123"
    with ossfs.open(file, "wb") as f_wb:
        f_wb.write(data)
    with ossfs.open(file) as f_r, io.BytesIO(data) as bytes_io:
        for fromossfs, fromio in zip(f_r, bytes_io):
            assert fromossfs == fromio
        f_r.seek(0)
        assert f_r.readline() == b"abc\n"
        assert f_r.readline() == b"123"
        f_r.seek(1)
        assert f_r.readline() == b"bc\n"

    with ossfs.open(file) as f_r:
        out = list(f_r)
    with ossfs.open(file) as f_r:
        out2 = f_r.readlines()
    assert out == out2
    assert b"".join(out) == data


@pytest.mark.parametrize("ossfs", ["async", "sync"], indirect=True)
def test_file_status(ossfs: Union["OSSFileSystem", "AioOSSFileSystem"], test_path: str):
    file = f"{test_path}/{function_name(ossfs)}/file"
    with ossfs.open(file, "wb") as f_wb:
        assert not f_wb.readable()
        assert not f_wb.seekable()
        assert f_wb.writable()

    with ossfs.open(file, "rb") as f_rb:
        assert f_rb.readable()
        assert f_rb.seekable()
        assert not f_rb.writable()


@pytest.mark.parametrize("ossfs", ["async", "sync"], indirect=True)
@pytest.mark.parametrize("data_size", [0, 20, 10 * 2**20])
@pytest.mark.parametrize("append_size", [0, 20, 10 * 2**20])
def test_append(
    ossfs: Union["OSSFileSystem", "AioOSSFileSystem"],
    test_path: str,
    data_size: int,
    append_size: int,
):
    file = f"{test_path}/{function_name(ossfs)}/file_{data_size}_{append_size}"
    data = os.urandom(data_size)
    extra = os.urandom(append_size)
    with ossfs.open(file, "wb") as f_wb:
        f_wb.write(data)
    assert ossfs.cat(file) == data
    with ossfs.open(file, "ab") as f_ab:
        f_ab.write(extra)  # append, write, small file
    assert ossfs.cat(file) == data + extra


@pytest.mark.parametrize("ossfs", ["async", "sync"], indirect=True)
def test_bigger_than_block_read(
    ossfs: Union["OSSFileSystem", "AioOSSFileSystem"], number_file: str
):
    with ossfs.open(number_file, "rb", block_size=3) as f_rb:
        out = []
        while True:
            data = f_rb.read(4)
            out.append(data)
            if len(data) == 0:
                break
    assert b"".join(out) == NUMBERS


@pytest.mark.parametrize("ossfs", ["async", "sync"], indirect=True)
def test_text_io__stream_wrapper_works(
    ossfs: Union["OSSFileSystem", "AioOSSFileSystem"], test_path: str
):
    """Ensure using TextIOWrapper works."""
    file = f"{test_path}/{function_name(ossfs)}/file"
    with ossfs.open(file, "wb") as f_wb:
        f_wb.write("\u00af\\_(\u30c4)_/\u00af".encode("utf-16-le"))

    with ossfs.open(file, "rb") as f_rb:
        with io.TextIOWrapper(f_rb, "utf-16-le") as stream:
            assert stream.readline() == "\u00af\\_(\u30c4)_/\u00af"


@pytest.mark.parametrize("ossfs", ["async", "sync"], indirect=True)
def test_text_io__basic(
    ossfs: Union["OSSFileSystem", "AioOSSFileSystem"], test_path: str
):
    """Text mode is now allowed."""
    file = f"{test_path}/{function_name(ossfs)}/file"
    with ossfs.open(file, "w", encoding="utf-8") as f_w:
        f_w.write("\u00af\\_(\u30c4)_/\u00af")

    with ossfs.open(file, "r", encoding="utf-8") as f_r:
        assert f_r.read() == "\u00af\\_(\u30c4)_/\u00af"


@pytest.mark.parametrize("ossfs", ["async", "sync"], indirect=True)
def test_text_io__override_encoding(
    ossfs: Union["OSSFileSystem", "AioOSSFileSystem"], test_path: str
):
    """Allow overriding the default text encoding."""
    file = f"{test_path}/{function_name(ossfs)}/file"

    with ossfs.open(file, "w", encoding="ibm500") as f_w:
        f_w.write("Hello, World!")

    with ossfs.open(file, "r", encoding="ibm500") as f_r:
        assert f_r.read() == "Hello, World!"


@pytest.mark.parametrize("ossfs", ["async", "sync"], indirect=True)
def test_readinto(ossfs: Union["OSSFileSystem", "AioOSSFileSystem"], test_path: str):
    file = f"{test_path}/{function_name(ossfs)}/file"

    with ossfs.open(file, "wb") as f_wb:
        f_wb.write(b"Hello, World!")

    contents = bytearray(15)

    with ossfs.open(file, "rb") as f_wb:
        assert f_wb.readinto(contents) == 13

    assert contents.startswith(b"Hello, World!")


@pytest.mark.parametrize("ossfs", ["async", "sync"], indirect=True)
def test_seek_reads(ossfs: Union["OSSFileSystem", "AioOSSFileSystem"], test_path: str):
    file = f"{test_path}/{function_name(ossfs)}/file"
    with ossfs.open(file, "wb") as f_wb:
        f_wb.write(os.urandom(5627146))
    with ossfs.open(file, "rb", blocksize=100) as f_rb:
        f_rb.seek(5561610)
        f_rb.read(65536)

        f_rb.seek(4)
        size = 562198
        data_2 = f_rb.read(size)
        assert len(data_2) == size

        f_rb.seek(562288)
        size = 562187
        data_3 = f_rb.read(size)
        assert len(data_3) == size



"""
Code of OSSFileSystem
"""

import copy
import logging
import os
from datetime import datetime
from hashlib import sha256
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Union

import oss2
from oss2.auth import AnonymousAuth
from oss2.models import PartInfo

from .base import DEFAULT_BLOCK_SIZE, SIMPLE_TRANSFER_THRESHOLD, BaseOSSFileSystem
from .exceptions import translate_oss_error
from .utils import as_progress_handler, prettify_info_result

if TYPE_CHECKING:
    from oss2.models import (
        GetObjectResult,
        InitMultipartUploadResult,
        PutObjectResult,
        SimplifiedObjectInfo,
    )


logger = logging.getLogger("ossfs")


class OSSFileSystem(BaseOSSFileSystem):
    """
    A pythonic file-systems interface to OSS (Object Storage Service)

    Examples
    --------
    >>> ossfs = OSSFileSystem(anon=False)
    >>> ossfs.ls('my-bucket/')
    ['my-file.txt']

    >>> with ossfs.open('my-bucket/my-file.txt', mode='rb') as f:
    ...     print(f.read())
    b'Hello, world!'
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._session = oss2.Session()

    def _get_bucket(
        self, bucket_name: str, connect_timeout: Optional[int] = None
    ) -> oss2.Bucket:
        """
        get the new bucket instance
        """
        if not self._endpoint:
            raise ValueError("endpoint is required")
        try:
            return oss2.Bucket(
                self._auth,
                self._endpoint,
                bucket_name,
                session=self._session,
                connect_timeout=connect_timeout,
                app_name="ossfs",
            )
        except oss2.exceptions.ClientError as err:
            raise FileNotFoundError(bucket_name) from err

    def _call_oss(
        self,
        method_name: str,
        *args,
        bucket: Optional[str] = None,
        timeout: Optional[int] = None,
        retry: int = 3,
        **kwargs,
    ):
        if bucket:
            service = self._get_bucket(bucket, timeout)
        else:
            service = oss2.Service(
                self._auth,
                endpoint=self._endpoint,
                connect_timeout=timeout,
            )
        for count in range(retry):
            try:
                method = getattr(service, method_name, None)
                if not method:
                    method = getattr(oss2, method_name)
                    logger.debug("CALL: %s - %s - %s", method.__name__, args, kwargs)
                    out = method(service, *args, **kwargs)
                else:
                    logger.debug("CALL: %s - %s - %s", method.__name__, args, kwargs)
                    out = method(*args, **kwargs)
                return out
            except oss2.exceptions.RequestError as err:  # noqa: PERF203
                logger.debug("Retryable error: %s, try %s times", err, count + 1)
                error = err
            except oss2.exceptions.OssError as err:
                logger.debug("Nonretryable error: %s", err)
                error = err
                break
        raise translate_oss_error(error) from error

    def _ls_bucket(self, connect_timeout: Optional[int]) -> List[Dict[str, Any]]:
        if "" not in self.dircache:
            results: List[Dict[str, Any]] = []
            if isinstance(self._auth, AnonymousAuth):
                logging.warning("cannot list buckets if not logged in")
                return []
            try:
                for bucket in self._call_oss("BucketIterator", timeout=connect_timeout):
                    result = {
                        "name": bucket.name,
                        "type": "directory",
                        "size": 0,
                        "CreateTime": bucket.creation_date,
                    }
                    results.append(result)
            except oss2.exceptions.ClientError:
                pass
            self.dircache[""] = copy.deepcopy(results)
        else:
            results = self.dircache[""]
        return results

    def _get_object_info_list(
        self,
        bucket_name: str,
        prefix: str,
        delimiter: str,
        connect_timeout: Optional[int],
    ):
        """
        Wrap oss2.ObjectIterator return values into a
        fsspec form of file info
        """
        result = []
        obj: SimplifiedObjectInfo
        for obj in self._call_oss(
            "ObjectIterator",
            prefix=prefix,
            delimiter=delimiter,
            bucket=bucket_name,
            timeout=connect_timeout,
        ):
            data = self._transfer_object_info_to_dict(bucket_name, obj)
            result.append(data)
        return result

    def _ls_dir(
        self,
        path: str,
        delimiter: str = "/",
        refresh: bool = False,
        prefix: str = "",
        connect_timeout: Optional[int] = None,
        **kwargs,
    ) -> List[Dict]:
        norm_path = path.strip("/")
        if norm_path in self.dircache and not refresh and not prefix and delimiter:
            return self.dircache[norm_path]

        logger.debug("Get directory listing page for %s", norm_path)
        bucket_name, key = self.split_path(norm_path)
        if not delimiter or prefix:
            if key:
                prefix = f"{key}/{prefix}"
        else:
            if norm_path in self.dircache and not refresh:
                return self.dircache[norm_path]
            if key:
                prefix = f"{key}/"

        try:
            self.dircache[norm_path] = self._get_object_info_list(
                bucket_name, prefix, delimiter, connect_timeout
            )
            return self.dircache[norm_path]
        except oss2.exceptions.AccessDenied:
            return []

    @prettify_info_result
    def ls(self, path: str, detail: bool = True, **kwargs):  # noqa: ARG002
        connect_timeout = kwargs.pop("connect_timeout", 60)
        norm_path = self._strip_protocol(path).strip("/")
        if norm_path == "":
            return self._ls_bucket(connect_timeout)
        files = self._ls_dir(path, connect_timeout=connect_timeout)
        if not files and "/" in norm_path:
            files = self._ls_dir(self._parent(path), connect_timeout=connect_timeout)
            files = [
                file
                for file in files
                if file["type"] != "directory" and file["name"].strip("/") == norm_path
            ]

        return files

    @prettify_info_result
    def find(
        self,
        path: str,
        maxdepth: Optional[int] = None,
        withdirs: bool = False,
        detail: bool = False,  # noqa: ARG002
        **kwargs,
    ):
        """List all files below path.

        Like posix ``find`` command without conditions

        Parameters
        ----------
        path : str
        maxdepth: int or None
            If not None, the maximum number of levels to descend
        withdirs: bool
            Whether to include directory paths in the output. This is True
            when used by glob, but users usually only want files.
        kwargs are passed to ``ls``.
        """
        out = {}
        prefix = kwargs.pop("prefix", "")
        path = self._verify_find_arguments(path, maxdepth, withdirs, prefix)
        if prefix:
            connect_timeout = kwargs.get("connect_timeout")
            for info in self._ls_dir(
                path, delimiter="", prefix=prefix, connect_timeout=connect_timeout
            ):
                out.update({info["name"]: info})
        else:
            for _, dirs, files in self.walk(path, maxdepth, detail=True, **kwargs):
                if withdirs:
                    files.update(dirs)
                out.update({info["name"]: info for name, info in files.items()})
            if self.isfile(path) and path not in out:
                # walk works on directories, but find should also return [path]
                # when path happens to be a file
                out[path] = {}
        names = sorted(out)
        return {name: out[name] for name in names}

    def _directory_exists(self, dirname: str, **kwargs):
        connect_timeout = kwargs.pop("connect_timeout", None)
        ls_result = self._ls_dir(dirname, connect_timeout=connect_timeout)
        return bool(ls_result)

    def _bucket_exist(self, bucket_name: str):
        if not bucket_name:
            return False
        try:
            self._call_oss("get_bucket_info", bucket=bucket_name)
        except (oss2.exceptions.OssError, PermissionError):
            return False
        return True

    def exists(self, path: str, **kwargs) -> bool:
        """Is there a file at the given path"""
        norm_path = self._strip_protocol(path).lstrip("/")
        if norm_path == "":
            return True

        bucket_name, obj_name = self.split_path(path)

        if not self._bucket_exist(bucket_name):
            return False

        connect_timeout = kwargs.get("connect_timeout")
        if not obj_name:
            return True

        if self._call_oss(
            "object_exists",
            obj_name,
            bucket=bucket_name,
            timeout=connect_timeout,
        ):
            return True

        return self._directory_exists(path, **kwargs)

    def ukey(self, path: str):
        """Hash of file properties, to tell if it has changed"""
        bucket_name, obj_name = self.split_path(path)
        obj_stream = self._call_oss("get_object", obj_name, bucket=bucket_name)
        return obj_stream.server_crc

    def checksum(self, path: str):
        """Unique value for current version of file

        If the checksum is the same from one moment to another, the contents
        are guaranteed to be the same. If the checksum changes, the contents
        *might* have changed.

        This should normally be overridden; default will probably capture
        creation/modification timestamp (which would be good) or maybe
        access timestamp (which would be bad)
        """
        return sha256(
            (str(self.ukey(path)) + str(self.info(path))).encode()
        ).hexdigest()

    def cp_file(self, path1: str, path2: str, **kwargs):
        """
        Copy within two locations in the filesystem
        # todo: big file optimization
        """
        bucket_name1, obj_name1 = self.split_path(path1)
        bucket_name2, obj_name2 = self.split_path(path2)
        self.invalidate_cache(self._parent(path2))
        if bucket_name1 != bucket_name2:
            tempdir = "." + self.ukey(path1)
            self.get_file(path1, tempdir, **kwargs)
            self.put_file(tempdir, path2, **kwargs)
            os.remove(tempdir)
        else:
            connect_timeout = kwargs.pop("connect_timeout", None)
            self._call_oss(
                "copy_object",
                bucket_name1,
                obj_name1,
                obj_name2,
                bucket=bucket_name1,
                timeout=connect_timeout,
            )

    def _rm(self, path: Union[str, List[str]]):
        """Delete files.

        Parameters
        ----------
        path: str or list of str
            File(s) to delete.
        """
        if isinstance(path, list):
            for file in path:
                self._rm(file)
            return
        bucket_name, obj_name = self.split_path(path)
        self.invalidate_cache(self._parent(path))
        self._call_oss("delete_object", obj_name, bucket=bucket_name)

    def _bulk_delete(self, pathlist, **kwargs):
        """
        Remove multiple keys with one call

        Parameters
        ----------
        pathlist : list(str)
            The keys to remove, must all be in the same bucket.
            Must have 0 < len <= 1000
        """
        if not pathlist:
            return
        bucket, key_list = self._get_batch_delete_key_list(pathlist)
        self._call_oss("batch_delete_objects", key_list, bucket=bucket)

    def rm(self, path: Union[str, List[str]], recursive=False, maxdepth=None):
        """Delete files.

        Parameters
        ----------
        path: str or list of str
            File(s) to delete.
        recursive: bool
            If file(s) are directories, recursively delete contents and then
            also remove the directory
        maxdepth: int or None
            Depth to pass to walk for finding files to delete, if recursive.
            If None, there will be no limit and infinite recursion may be
            possible.
        """

        if isinstance(path, list):
            for file in path:
                self.rm(file)
            return

        path_expand = self.expand_path(path, recursive=recursive, maxdepth=maxdepth)

        def chunks(lst: list, num: int):
            for i in range(0, len(lst), num):
                yield lst[i : i + num]

        for files in chunks(path_expand, 1000):
            self._bulk_delete(files)

    def get_path(self, rpath: str, lpath: str, **kwargs):
        """
        Copy single remote path to local
        """
        if self.isdir(rpath):
            os.makedirs(lpath, exist_ok=True)
        else:
            self.get_file(rpath, lpath, **kwargs)

    def get_file(
        self, rpath: str, lpath: str, callback: Optional[Callable] = None, **kwargs
    ):
        """
        Copy single remote file to local
        """
        bucket_name, obj_name = self.split_path(rpath)
        kwargs.setdefault("progress_callback", as_progress_handler(callback))
        if self.isdir(rpath):
            os.makedirs(lpath, exist_ok=True)
            return
        connect_timeout = kwargs.pop("connect_timeout", None)
        bucket = self._get_bucket(bucket_name, connect_timeout)
        if self.size(rpath) >= SIMPLE_TRANSFER_THRESHOLD:
            oss2.resumable_download(bucket, obj_name, lpath, **kwargs)
        else:
            self._call_oss(
                "get_object_to_file",
                obj_name,
                lpath,
                bucket=bucket_name,
                timeout=connect_timeout,
                **kwargs,
            )

    def put_file(
        self, lpath: str, rpath: str, callback: Optional[Callable] = None, **kwargs
    ):
        """
        Copy single file to remote
        """
        kwargs.setdefault("progress_callback", as_progress_handler(callback))
        bucket_name, obj_name = self.split_path(rpath)
        if os.path.isdir(lpath):
            if obj_name:
                # don't make remote "directory"
                return
            self.mkdir(lpath)
        else:
            connect_timeout = kwargs.pop("connect_timeout", None)
            bucket = self._get_bucket(bucket_name, connect_timeout)
            if os.path.getsize(lpath) >= SIMPLE_TRANSFER_THRESHOLD:
                oss2.resumable_upload(bucket, obj_name, lpath, **kwargs)
            else:
                self._call_oss(
                    "put_object_from_file",
                    obj_name,
                    lpath,
                    bucket=bucket_name,
                    timeout=connect_timeout,
                    **kwargs,
                )
        self.invalidate_cache(self._parent(rpath))

    def created(self, path: str):
        """Return the created timestamp of a file as a datetime.datetime"""
        bucket_name, obj_name = self.split_path(path)
        if obj_name:
            raise NotImplementedError("OSS has no created timestamp")
        bucket_info = self._call_oss("get_bucket_info", bucket=bucket_name)
        timestamp = bucket_info.creation_date
        return datetime.fromtimestamp(timestamp)  # noqa: DTZ006

    def modified(self, path: str):
        """Return the modified timestamp of a file as a datetime.datetime"""
        bucket_name, obj_name = self.split_path(path)
        if not obj_name or self.isdir(path):
            raise NotImplementedError("bucket has no modified timestamp")
        simplifiedmeta = self._call_oss("get_object_meta", obj_name, bucket=bucket_name)
        return int(
            datetime.strptime(  # noqa: DTZ007
                simplifiedmeta.headers["Last-Modified"],
                "%a, %d %b %Y %H:%M:%S %Z",
            ).timestamp()
        )

    def append_object(self, path: str, location: int, value: bytes) -> int:
        """
        Append bytes to the object
        """
        bucket_name, obj_name = self.split_path(path)
        result = self._call_oss(
            "append_object",
            obj_name,
            location,
            value,
            bucket=bucket_name,
        )
        return result.next_position

    def get_object(self, path: str, start: int, end: int) -> bytes:
        """
        Return object bytes in range
        """
        headers = {"x-oss-range-behavior": "standard"}
        bucket_name, obj_name = self.split_path(path)
        try:
            object_stream = self._call_oss(
                "get_object",
                obj_name,
                bucket=bucket_name,
                byte_range=(start, end),
                headers=headers,
            )
        except oss2.exceptions.ServerError:  # noqa: TRY203
            raise
        return object_stream.read()

    def sign(self, path: str, expiration: int = 100, **kwargs):
        raise NotImplementedError("Sign is not implemented for this filesystem")

    def pipe_file(self, path: str, value: str, **kwargs):
        """Set the bytes of given file"""
        bucket, key = self.split_path(path)
        block_size = kwargs.get("block_size", DEFAULT_BLOCK_SIZE)
        # 5 GB is the limit for an OSS PUT
        self.invalidate_cache(path)
        if len(value) < min(5 * 2**30, 2 * block_size):
            self._call_oss("put_object", key, value, bucket=bucket, **kwargs)
            return
        mpu: InitMultipartUploadResult = self._call_oss(
            "init_multipart_upload", key, bucket=bucket, **kwargs
        )
        parts: List[PartInfo] = []
        for i, off in enumerate(range(0, len(value), block_size)):
            data = value[off : off + block_size]
            part_number = i + 1
            out: PutObjectResult = self._call_oss(
                "upload_part",
                key,
                mpu.upload_id,
                part_number,
                data,
                bucket=bucket,
            )
            parts.append(
                PartInfo(
                    part_number,
                    out.etag,
                    size=len(data),
                    part_crc=out.crc,
                )
            )
        self._call_oss(
            "complete_multipart_upload",
            key,
            mpu.upload_id,
            parts,
            bucket=bucket,
        )

    @prettify_info_result
    def info(self, path, **kwargs):
        norm_path = self._strip_protocol(path).lstrip("/")
        if norm_path == "":
            result = {"name": path, "size": 0, "type": "directory"}
        else:
            result = super().info(path, **kwargs)
        if "StorageClass" in result:
            del result["StorageClass"]
        if "CreateTime" in result:
            del result["CreateTime"]
        return result

    def cat_file(
        self,
        path: str,
        start: Optional[int] = None,
        end: Optional[int] = None,
        **kwargs,
    ):
        bucket, object_name = self.split_path(path)
        object_stream: GetObjectResult = self._call_oss(
            "get_object",
            bucket=bucket,
            key=object_name,
            byte_range=(start, end),
            **kwargs,
        )

        return object_stream.read()



"""
Test all oss errors
"""

from typing import TYPE_CHECKING, Union

import pytest

from tests.conftest import function_name

if TYPE_CHECKING:
    from oss2 import Bucket

    from ossfs import AioOSSFileSystem, OSSFileSystem


@pytest.mark.parametrize("ossfs", ["sync", "async"], indirect=True)
def test_errors(
    ossfs: Union["OSSFileSystem", "AioOSSFileSystem"], test_path: str, bucket: "Bucket"
):
    path = f"{test_path}/{function_name(ossfs)}/"

    with pytest.raises(FileNotFoundError):
        ossfs.open(path + "none", "rb")

    with pytest.raises(FileNotFoundError):
        ossfs.mv(path + "x", path + "y")

    with pytest.raises(FileNotFoundError):
        ossfs.open("x", "rb")

    with pytest.raises(FileNotFoundError):
        ossfs.open("xxxx", "rb")

    with pytest.raises(PermissionError):
        ossfs.rm("/non-exist-bucket")

    with pytest.raises(ValueError), ossfs.open(path + "temp", "wb") as file_obj:
        file_obj.read()

    bucket.put_object(path.split("/", 2)[-1] + "temp", "foobar")
    with pytest.raises(ValueError):  # noqa: PT012
        file_obj = ossfs.open(path + "temp", "rb")
        file_obj.close()
        file_obj.read()



"""tests for bucket level operations"""

from typing import TYPE_CHECKING, Union

import pytest
from oss2.auth import AnonymousAuth

if TYPE_CHECKING:
    from ossfs import AioOSSFileSystem, OSSFileSystem


@pytest.mark.parametrize("ossfs", ["sync", "async"], indirect=True)
def test_ls_bucket(
    ossfs: Union["OSSFileSystem", "AioOSSFileSystem"],
    test_bucket_name: str,
    anonymous_bucket_name: str,
    caplog,
):
    bucket_infos = ossfs.ls("/")
    bucket_info_dict = {}
    for bucket_info in bucket_infos:
        bucket_info_dict[bucket_info["name"]] = bucket_info
    assert f"/{test_bucket_name}" in bucket_info_dict
    assert f"/{anonymous_bucket_name}" in bucket_info_dict

    auth = ossfs._auth
    try:
        ossfs._auth = AnonymousAuth()
        ossfs.invalidate_cache()
        bucket_infos = ossfs.ls("/")
        assert bucket_infos == []
        assert "cannot list buckets if not logged in" in caplog.text
    finally:
        ossfs._auth = auth


@pytest.mark.parametrize("ossfs", ["sync", "async"], indirect=True)
def test_bucket_is_dir(
    ossfs: Union["OSSFileSystem", "AioOSSFileSystem"],
    test_bucket_name: str,
    anonymous_bucket_name: str,
):
    assert ossfs.isdir("")
    assert ossfs.isdir("/")
    assert ossfs.isdir(test_bucket_name)
    assert not ossfs.isdir("non-exist-bucket")
    assert ossfs.isdir(f"/{anonymous_bucket_name}")


@pytest.mark.parametrize("ossfs", ["sync", "async"], indirect=True)
def test_bucket_is_file(
    ossfs: Union["OSSFileSystem", "AioOSSFileSystem"],
    test_bucket_name: str,
    anonymous_bucket_name: str,
):
    assert not ossfs.isfile("")
    assert not ossfs.isfile("/")
    assert not ossfs.isfile(test_bucket_name)
    assert not ossfs.isfile("non-exist-bucket")
    assert not ossfs.isfile(f"/{anonymous_bucket_name}")


@pytest.mark.parametrize("ossfs", ["sync", "async"], indirect=True)
def test_bucket_exists(
    ossfs: Union["OSSFileSystem", "AioOSSFileSystem"],
    test_bucket_name: str,
    anonymous_bucket_name: str,
):
    assert ossfs.exists("")
    assert ossfs.exists("/")
    assert ossfs.exists(test_bucket_name)
    assert not ossfs.exists("non-exist-bucket")
    assert ossfs.exists(f"/{anonymous_bucket_name}")


@pytest.mark.parametrize("ossfs", ["sync", "async"], indirect=True)
def test_bucket_info(
    ossfs: Union["OSSFileSystem", "AioOSSFileSystem"],
    test_bucket_name: str,
    anonymous_bucket_name: str,
):
    assert ossfs.info("") == {
        "Key": "",
        "Size": 0,
        "type": "directory",
        "size": 0,
        "name": "",
    }
    assert ossfs.info("/") == {
        "Key": "/",
        "Size": 0,
        "type": "directory",
        "size": 0,
        "name": "/",
    }
    assert ossfs.info(test_bucket_name) == {
        "Key": test_bucket_name,
        "type": "directory",
        "Size": 0,
        "size": 0,
        "name": test_bucket_name,
    }
    with pytest.raises(FileNotFoundError):
        assert ossfs.info("non-exist-bucket")
    assert ossfs.info(f"/{anonymous_bucket_name}") == {
        "Key": f"/{anonymous_bucket_name}",
        "type": "directory",
        "Size": 0,
        "size": 0,
        "name": f"/{anonymous_bucket_name}",
    }



"""
Three different login test
Sts, anonymous, accesskey
"""

import json
import os
from typing import Union

import oss2
import pytest
from aliyunsdkcore import client
from aliyunsdksts.request.v20150401 import AssumeRoleRequest

from ossfs import AioOSSFileSystem, OSSFileSystem

STSAccessKeyId = os.getenv("OSS_TEST_STS_ID", "")
STSAccessKeySecret = os.getenv("OSS_TEST_STS_KEY", "")
STSArn = os.getenv("OSS_TEST_STS_ARN", "")


def fetch_sts_token(access_key_id: str, access_key_secret: str, role_arn: str):
    clt = client.AcsClient(access_key_id, access_key_secret, "cn-hangzhou")
    req = AssumeRoleRequest.AssumeRoleRequest()

    req.set_accept_format("json")
    req.set_RoleArn(role_arn)
    req.set_RoleSessionName("oss-python-sdk-example")

    body = clt.do_action_with_exception(req)

    j = json.loads(oss2.to_unicode(body))

    access_key_id = j["Credentials"]["AccessKeyId"]
    access_key_secret = j["Credentials"]["AccessKeySecret"]
    security_token = j["Credentials"]["SecurityToken"]

    return access_key_id, access_key_secret, security_token


@pytest.mark.parametrize("ossfs", ["sync", "async"], indirect=True)
def test_access_key_login(
    ossfs: Union["OSSFileSystem", "AioOSSFileSystem"], test_bucket_name: str
):
    ossfs.ls(test_bucket_name)


@pytest.mark.parametrize("aio", [False, True])
def test_sts_login(endpoint: str, test_bucket_name: str, aio: bool):
    key, secret, token = fetch_sts_token(STSAccessKeyId, STSAccessKeySecret, STSArn)
    kwargs = {"key": key, "secret": secret, "token": token, "endpoint": endpoint}
    if aio:  # noqa: SIM108
        ossfs = AioOSSFileSystem(**kwargs)
    else:
        ossfs = OSSFileSystem(**kwargs)
    ossfs.ls(test_bucket_name)


@pytest.mark.parametrize("aio", [False, True])
def test_set_endpoint(endpoint: str, test_bucket_name: str, monkeypatch, aio: bool):
    key, secret, token = fetch_sts_token(STSAccessKeyId, STSAccessKeySecret, STSArn)
    monkeypatch.delenv("OSS_ENDPOINT")
    kwargs = {"key": key, "secret": secret, "token": token, "endpoint": None}
    if aio:  # noqa: SIM108
        ossfs = AioOSSFileSystem(**kwargs)
    else:
        ossfs = OSSFileSystem(**kwargs)
    with pytest.raises(ValueError):
        ossfs.ls(test_bucket_name)
    ossfs.set_endpoint(endpoint)
    ossfs.ls(test_bucket_name)


@pytest.mark.parametrize("aio", [False, True])
def test_env_endpoint(endpoint: str, test_bucket_name: str, monkeypatch, aio: bool):
    key, secret, token = fetch_sts_token(STSAccessKeyId, STSAccessKeySecret, STSArn)
    monkeypatch.setenv("OSS_ENDPOINT", endpoint)
    kwargs = {"key": key, "secret": secret, "token": token, "endpoint": None}
    if aio:  # noqa: SIM108
        ossfs = AioOSSFileSystem(**kwargs)
    else:
        ossfs = OSSFileSystem(**kwargs)
    ossfs.ls(test_bucket_name)


@pytest.mark.parametrize("aio", [False, True])
def test_anonymous_login(file_in_anonymous: str, endpoint: str, aio: bool):
    if aio:
        ossfs = AioOSSFileSystem(endpoint=endpoint)
    else:
        ossfs = OSSFileSystem(endpoint=endpoint)
    ossfs.get_object(file_in_anonymous, 1, 100)



"""
Test all oss object related methods
"""

import os
import time
from typing import TYPE_CHECKING, Union

import fsspec
import pytest

from tests.conftest import NUMBERS, bucket_relative_path, function_name

if TYPE_CHECKING:
    from oss2 import Bucket

    from ossfs import AioOSSFileSystem, OSSFileSystem


@pytest.mark.parametrize("ossfs", ["sync", "async"], indirect=True)
def test_info(
    ossfs: Union["OSSFileSystem", "AioOSSFileSystem"], test_path: str, bucket: "Bucket"
):
    # test file info
    object_name_foo = f"{test_path}/{function_name(ossfs)}/foo"
    bucket.put_object(bucket_relative_path(object_name_foo), "foo")
    info = ossfs.info(object_name_foo)
    linfo = ossfs.ls(object_name_foo, detail=True)[0]

    assert abs(info.pop("LastModified") - linfo.pop("LastModified")) <= 1
    assert info["size"] == 3
    assert info == linfo

    # test not exist dir
    file_path = f"{test_path}/{function_name(ossfs)}/"
    object_name_bar = file_path + "bar"
    with pytest.raises(FileNotFoundError):
        ossfs.info(object_name_bar)
    ossfs.invalidate_cache(file_path)

    # add a new file then we can info this new dir.
    bucket.put_object(bucket_relative_path(object_name_bar), "bar")
    ossfs.info(file_path + "bar")


@pytest.mark.parametrize("ossfs", ["sync", "async"], indirect=True)
def test_checksum(
    ossfs: Union["OSSFileSystem", "AioOSSFileSystem"], test_path: str, bucket: "Bucket"
):
    # don't make local "directory"
    object_name = f"{test_path}/{function_name(ossfs)}"

    # change one file, using cache
    bucket.put_object(bucket_relative_path(object_name), "foo")
    checksum = ossfs.checksum(object_name)

    # Test changed contents
    bucket.put_object(bucket_relative_path(object_name), "bar")
    assert checksum != ossfs.checksum(object_name)
    # Test for same contents

    # sleep 1 second to make sure the last modificed changes
    time.sleep(1)
    bucket.put_object(bucket_relative_path(object_name), "foo")
    ossfs.invalidate_cache()
    assert checksum != ossfs.checksum(object_name)

    # Test for nonexistent file
    bucket.delete_object(bucket_relative_path(object_name))
    with pytest.raises(FileNotFoundError):
        ossfs.checksum(object_name)


@pytest.mark.parametrize("ossfs", ["sync", "async"], indirect=True)
def test_ls_object(
    # don't make local "directory"
    ossfs: Union["OSSFileSystem", "AioOSSFileSystem"],
    test_path: str,
    bucket: "Bucket",
):
    path = f"{test_path}/{function_name(ossfs)}/"
    assert ossfs.ls(path + "nonexistent") == []
    ossfs.invalidate_cache(path)
    filename = path + "accounts.1.json"
    bucket.put_object(bucket_relative_path(filename), "")
    assert filename in ossfs.ls(path, detail=False)


@pytest.mark.parametrize("ossfs", ["sync", "async"], indirect=True)
def test_ls_dir(
    ossfs: Union["OSSFileSystem", "AioOSSFileSystem"], test_path: str, bucket: "Bucket"
):
    path = f"{test_path}/{function_name(ossfs)}/"
    file = path + "file"
    bucket.put_object(bucket_relative_path(file), "")
    assert file in ossfs.ls(path, detail=False)


@pytest.mark.parametrize("ossfs", ["sync", "async"], indirect=True)
def test_ls_and_touch(
    ossfs: Union["OSSFileSystem", "AioOSSFileSystem"], test_path: str, bucket: "Bucket"
):
    # don't make local "directory"
    path = f"{test_path}/{function_name(ossfs)}/"
    file_a = path + "a"
    file_b = path + "b"
    assert not bucket.object_exists(bucket_relative_path(path))
    ossfs.touch(file_a)
    ossfs.touch(file_b)
    ls_result = ossfs.ls(path, detail=True)
    assert {result["Key"] for result in ls_result} == {file_a, file_b}
    ls_result = ossfs.ls(path, detail=False)
    assert set(ls_result) == {file_a, file_b}


@pytest.mark.parametrize("ossfs", ["sync", "async"], indirect=True)
def test_isfile(
    # don't make local "directory"
    ossfs: Union["OSSFileSystem", "AioOSSFileSystem"],
    test_path: str,
    bucket: "Bucket",
):
    "test isfile in ossfs"
    path = f"{test_path}/{function_name(ossfs)}/"
    file_foo = path + "foo"
    assert not ossfs.isfile("")
    assert not ossfs.isfile("/")
    assert not ossfs.isfile(path)

    assert not ossfs.isfile(file_foo)
    ossfs.invalidate_cache()
    bucket.put_object(bucket_relative_path(file_foo), "foo")
    assert ossfs.isfile(file_foo)


@pytest.mark.parametrize("ossfs", ["sync", "async"], indirect=True)
def test_isdir(
    ossfs: Union["OSSFileSystem", "AioOSSFileSystem"], test_path: str, bucket: "Bucket"
):
    "test isdir in ossfs"
    path = f"{test_path}/{function_name(ossfs)}/"
    file_foo = path + "foo"
    bucket.put_object(bucket_relative_path(file_foo), "foo")
    assert not ossfs.isdir(file_foo)
    assert ossfs.isdir(test_path)
    assert ossfs.isdir(path)
    assert ossfs.isdir(path.rstrip("/"))


@pytest.mark.parametrize("ossfs", ["sync", "async"], indirect=True)
def test_rm(
    ossfs: Union["OSSFileSystem", "AioOSSFileSystem"], test_path: str, bucket: "Bucket"
):
    path = f"{test_path}/{function_name(ossfs)}/"
    file_foo = path + "foo"
    nest_file = path + "nested/nested2/file1"
    bucket.put_object(bucket_relative_path(file_foo), "foo")
    assert bucket.object_exists(bucket_relative_path(file_foo))
    ossfs.rm(file_foo)
    assert not bucket.object_exists(bucket_relative_path(file_foo))
    with pytest.raises(FileNotFoundError):
        ossfs.rm("nonexistent")
    bucket.put_object(bucket_relative_path(nest_file), "foo")
    ossfs.rm(path + "nested", recursive=True)
    assert not bucket.object_exists(bucket_relative_path(nest_file))


@pytest.mark.parametrize("ossfs", ["sync", "async"], indirect=True)
def test_bulk_delete(
    ossfs: Union["OSSFileSystem", "AioOSSFileSystem"], test_path: str, bucket: "Bucket"
):
    path = f"{test_path}/{function_name(ossfs)}/"
    nest_file1 = path + "nested/file1"
    nest_file2 = path + "nested/file2"
    bucket.put_object(bucket_relative_path(nest_file1), "foo")
    bucket.put_object(bucket_relative_path(nest_file2), "bar")
    filelist = ossfs.find(path + "nested/")
    assert set(filelist) == {nest_file1, nest_file2}
    ossfs.rm(filelist)
    assert not bucket.object_exists(bucket_relative_path(nest_file1))
    assert not bucket.object_exists(bucket_relative_path(nest_file2))


@pytest.mark.parametrize("ossfs", ["sync", "async"], indirect=True)
def test_ossfs_file_access(
    ossfs: Union["OSSFileSystem", "AioOSSFileSystem"], number_file: str
):
    assert ossfs.cat(number_file) == NUMBERS
    assert ossfs.head(number_file, 3) == NUMBERS[:3]
    assert ossfs.tail(number_file, 3) == NUMBERS[-3:]
    assert ossfs.tail(number_file, 10000) == NUMBERS
    assert ossfs.info(number_file)["Size"] == len(NUMBERS)


@pytest.mark.parametrize("ossfs", ["sync", "async"], indirect=True)
def test_du(
    ossfs: Union["OSSFileSystem", "AioOSSFileSystem"], test_path: str, bucket: "Bucket"
):
    path = f"{test_path}/{function_name(ossfs)}/"
    file1 = path + "file1"
    file2 = path + "dir/file2"
    file3 = path + "dir/file3"
    bucket.put_object(bucket_relative_path(file1), b"1234567890")
    bucket.put_object(bucket_relative_path(file2), b"12345")
    bucket.put_object(bucket_relative_path(file3), b"1234567890" * 2)

    du_result = ossfs.du(path, total=False)
    assert all(isinstance(v, int) and v >= 0 for v in du_result.values())
    assert du_result[file1] == 10
    assert du_result[file2] == 5
    assert du_result[file3] == 20
    assert ossfs.du(path, total=True) == 35


@pytest.mark.parametrize("ossfs", ["sync", "async"], indirect=True)
def test_ossfs_ls(
    ossfs: Union["OSSFileSystem", "AioOSSFileSystem"], test_path: str, bucket: "Bucket"
):
    path = f"{test_path}/{function_name(ossfs)}/"
    filename = path + "nested/file1"
    bucket.put_object(bucket_relative_path(filename), "foo")
    assert filename not in ossfs.ls(path, detail=False)
    assert filename in ossfs.ls(path + "nested/", detail=False)
    assert filename in ossfs.ls(path + "nested", detail=False)
    ls_result = ossfs.ls(path + "nested", detail=True)
    assert all(isinstance(item, dict) for item in ls_result)


@pytest.mark.parametrize("ossfs", ["sync", "async"], indirect=True)
def test_ossfs_big_ls(
    ossfs: Union["OSSFileSystem", "AioOSSFileSystem"], test_path: str, bucket: "Bucket"
):
    path = f"{test_path}/{function_name(ossfs)}/"
    for num in range(120):
        bucket.put_object(bucket_relative_path(f"{path}{num}.part"), "foo")
    assert len(ossfs.ls(path, detail=False, connect_timeout=600)) == 120


@pytest.mark.parametrize("ossfs", ["sync", "async"], indirect=True)
def test_ossfs_glob(
    ossfs: Union["OSSFileSystem", "AioOSSFileSystem"], test_path: str, bucket: "Bucket"
):
    path = f"{test_path}/{function_name(ossfs)}/"
    file1 = path + "nested/file.dat"
    file2 = path + "nested/filedat"
    bucket.put_object(bucket_relative_path(file1), "foo")
    bucket.put_object(bucket_relative_path(file2), "bar")
    assert file1 not in ossfs.glob(path + "")
    assert file1 not in ossfs.glob(path + "*")
    assert file1 not in ossfs.glob(path + "nested")
    assert file1 in ossfs.glob(path + "nested/*")
    assert file1 in ossfs.glob(path + "nested/file*")
    assert all(
        any(p.startswith(f + "/") or p == f for p in ossfs.find(path))
        for f in ossfs.glob(path + "nested/*")
    )
    out = ossfs.glob(path + "nested/*")
    assert {file1, file2} == set(out)

    # Make sure glob() deals with the dot character (.) correctly.
    assert file1 in ossfs.glob(path + "nested/file.*")
    assert file2 not in ossfs.glob(path + "nested/file.*")


@pytest.mark.parametrize("ossfs", ["sync", "async"], indirect=True)
def test_copy(
    ossfs: Union["OSSFileSystem", "AioOSSFileSystem"],
    number_file: str,
    test_path: str,
    bucket: "Bucket",
):
    path = f"{test_path}/{function_name(ossfs)}/"
    new_file = path + "file"
    ossfs.copy(number_file, new_file)
    assert bucket.get_object(bucket_relative_path(new_file)).read() == NUMBERS


@pytest.mark.parametrize("ossfs", ["sync", "async"], indirect=True)
def test_move(
    ossfs: Union["OSSFileSystem", "AioOSSFileSystem"],
    number_file: str,
    test_path: str,
    bucket: "Bucket",
):
    path = f"{test_path}/{function_name(ossfs)}/"
    from_file = path + "from"
    to_file = path + "to"
    ossfs.copy(number_file, from_file)
    assert bucket.object_exists(bucket_relative_path(from_file))
    assert not bucket.object_exists(bucket_relative_path(to_file))
    ossfs.mv(from_file, to_file)
    assert bucket.get_object(bucket_relative_path(to_file)).read() == NUMBERS
    assert not bucket.object_exists(bucket_relative_path(from_file))


@pytest.mark.parametrize("ossfs", ["sync", "async"], indirect=True)
@pytest.mark.parametrize("size", [2**10, 2**20, 10 * 2**20])
def test_get_put(
    ossfs: Union["OSSFileSystem", "AioOSSFileSystem"],
    tmpdir,
    test_path: str,
    size: int,
    bucket: "Bucket",
):
    path = f"{test_path}/{function_name(ossfs)}/"

    local_file = str(tmpdir.join("number"))
    data = os.urandom(size)
    with open(local_file, "wb") as f_write:
        f_write.write(data)

    remote_file = path + "file"
    ossfs.put(local_file, remote_file)
    assert bucket.object_exists(bucket_relative_path(remote_file))
    assert bucket.get_object(bucket_relative_path(remote_file)).read() == data

    get_file = str(tmpdir.join("get"))
    ossfs.get(remote_file, get_file)
    with open(get_file, "rb") as f_read:
        assert f_read.read() == data


@pytest.mark.parametrize("ossfs", ["sync", "async"], indirect=True)
@pytest.mark.parametrize("size", [2**10, 20 * 2**20])
def test_pipe_cat_big(
    ossfs: "OSSFileSystem", size: int, test_path: str, bucket: "Bucket"
):
    path = f"{test_path}/{function_name(ossfs)}/"
    bigfile = path + "bigfile"
    data = os.urandom(size)
    ossfs.pipe(bigfile, data)
    assert bucket.get_object(bucket_relative_path(bigfile)).read() == data


@pytest.mark.parametrize("ossfs", ["sync", "async"], indirect=True)
def test_touch(
    ossfs: Union["OSSFileSystem", "AioOSSFileSystem"], test_path: str, bucket: "Bucket"
):
    # create
    path = f"{test_path}/{function_name(ossfs)}/"
    filename = path + "file"
    assert not bucket.object_exists(bucket_relative_path(filename))
    ossfs.touch(filename)
    assert bucket.object_exists(bucket_relative_path(filename))
    assert bucket.get_object(bucket_relative_path(filename)).read() == b""

    bucket.put_object(bucket_relative_path(filename), b"data")
    with pytest.raises(NotImplementedError):
        ossfs.touch(filename, truncate=False)
    assert bucket.get_object(bucket_relative_path(filename)).read() == b"data"


@pytest.mark.parametrize("ossfs", ["sync", "async"], indirect=True)
def test_cat_missing(
    ossfs: Union["OSSFileSystem", "AioOSSFileSystem"], test_path: str, bucket: "Bucket"
):
    path = f"{test_path}/{function_name(ossfs)}/"
    file1 = path + "file0"
    file2 = path + "file1"
    ossfs.touch(file1)
    bucket.put_object(bucket_relative_path(file1), b"foo")
    with pytest.raises(FileNotFoundError):
        ossfs.cat([file1, file2], on_error="raise")
    out = ossfs.cat([file1, file2], on_error="omit")
    assert list(out) == [file1]
    out = ossfs.cat([file1, file2], on_error="return")
    assert file2 in out
    assert isinstance(out[file2], FileNotFoundError)


@pytest.mark.parametrize("ossfs", ["sync", "async"], indirect=True)
def test_get_directories(
    ossfs: Union["OSSFileSystem", "AioOSSFileSystem"],
    tmpdir,
    test_path: str,
    bucket: "Bucket",
):
    path = f"{test_path}/{function_name(ossfs)}/"
    bucket.put_object(bucket_relative_path(path + "dir/dirkey/key0"), "")
    bucket.put_object(bucket_relative_path(path + "dir/dirkey/key1"), "")
    bucket.put_object(bucket_relative_path(path + "dir/dir/key"), "")
    tmpdir_str = str(tmpdir)
    ossfs.get(path + "dir/", tmpdir_str, recursive=True)
    assert {"dirkey", "dir"} == set(os.listdir(tmpdir_str))
    assert os.listdir(os.path.join(tmpdir_str, "dir")) == ["key"]
    assert {"key0", "key1"} == set(os.listdir(os.path.join(tmpdir_str, "dirkey")))


@pytest.mark.parametrize("ossfs", ["sync", "async"], indirect=True)
def test_modified(
    ossfs: Union["OSSFileSystem", "AioOSSFileSystem"],
    test_path: str,
    test_bucket_name: str,
    bucket: "Bucket",
):
    path = f"{test_path}/{function_name(ossfs)}/"
    file_path = path + "file"
    notexist_path = path + "notexist"

    # Test file
    bucket.put_object(bucket_relative_path(file_path), "")
    modified = ossfs.modified(path=file_path)
    assert isinstance(modified, int)

    # Test file
    with pytest.raises(FileNotFoundError):
        modified = ossfs.modified(path=notexist_path)

    # Test directory
    with pytest.raises(NotImplementedError):
        modified = ossfs.modified(path=path)

    # Test bucket
    with pytest.raises(NotImplementedError):
        ossfs.modified(path=test_bucket_name)


@pytest.mark.parametrize("ossfs", ["sync", "async"], indirect=True)
def test_find_file_info_with_selector(
    ossfs: Union["OSSFileSystem", "AioOSSFileSystem"], test_path: str, bucket: "Bucket"
):
    path = f"{test_path}/{function_name(ossfs)}/"
    file_a = path + "test_file_a"
    file_b = path + "test_file_b"
    dir_a = path + "test_dir_a"
    file_c = path + "test_dir_a/test_file_c"

    bucket.put_object(bucket_relative_path(file_a), "")
    bucket.put_object(bucket_relative_path(file_b), "")
    bucket.put_object(bucket_relative_path(file_c), "")

    infos = ossfs.find(path, maxdepth=None, withdirs=True, detail=True)
    assert len(infos) == 4

    for info in infos.values():
        if (
            info["name"].endswith(file_a)
            or info["name"].endswith(file_b)
            or info["name"].endswith(file_c)
        ):
            assert info["type"] == "file"
        elif info["name"].rstrip("/").endswith(dir_a):
            assert info["type"] == "directory"
        else:
            raise ValueError(f"unexpected path {info['name']}")


@pytest.mark.parametrize("ossfs", ["sync", "async"], indirect=True)
def test_exists(
    ossfs: Union["OSSFileSystem", "AioOSSFileSystem"], test_path: str, bucket: "Bucket"
):
    path = f"{test_path}/{function_name(ossfs)}/"
    bucket.put_object(bucket_relative_path(path + "very/similar/prefix1"), "")
    bucket.put_object(bucket_relative_path(path + "very/similar/prefix2"), "")
    bucket.put_object(bucket_relative_path(path + "very/similar/prefix3/something"), "")

    assert not ossfs.exists(path + "very/similar/prefix")
    assert not ossfs.exists(path + "very/similar/prefi")
    assert not ossfs.exists(path + "very/similar/pref")

    assert ossfs.exists(path + "very/similar/")
    assert ossfs.exists(path + "very/similar/prefix1")
    assert ossfs.exists(path + "very/similar/prefix2")
    assert ossfs.exists(path + "very/similar/prefix3")
    assert ossfs.exists(path + "very/similar/prefix3/")
    assert ossfs.exists(path + "very/similar/prefix3/something")

    assert not ossfs.exists(path + "very/similar/prefix3/some")

    bucket.put_object(bucket_relative_path(path + "starting/very/similar/prefix"), "")

    assert not ossfs.exists(path + "starting/very/similar/prefix1")
    assert not ossfs.exists(path + "starting/very/similar/prefix3")
    assert not ossfs.exists(path + "starting/very/similar/prefix3/")
    assert not ossfs.exists(path + "starting/very/similar/prefix3/something")

    assert ossfs.exists(path + "starting/very/similar/prefix")
    assert not ossfs.exists(path + "starting/very/similar/prefix/")


@pytest.mark.parametrize("ossfs", ["sync", "async"], indirect=True)
def test_leading_forward_slash(
    ossfs: Union["OSSFileSystem", "AioOSSFileSystem"], test_path: str, bucket: "Bucket"
):
    path = f"{test_path}/{function_name(ossfs)}/"
    bucket.put_object(bucket_relative_path(path + "some/file"), "")
    assert ossfs.ls(path + "some/")
    path = path.lstrip("/")
    assert ossfs.exists(path + "some/file")
    assert ossfs.exists("/" + path + "some/file")


@pytest.mark.parametrize("ossfs", ["sync", "async"], indirect=True)
def test_find_with_prefix(
    ossfs: Union["OSSFileSystem", "AioOSSFileSystem"], test_path: str, bucket: "Bucket"
):
    path = f"{test_path}/{function_name(ossfs)}/"
    if not ossfs.exists(path):
        for cursor in range(100):
            bucket.put_object(
                bucket_relative_path(path + f"prefixes/test_{cursor}"), ""
            )
        bucket.put_object(bucket_relative_path(path + "prefixes2"), "")

    assert len(ossfs.find(path + "prefixes")) == 100
    assert len(ossfs.find(path, prefix="prefixes")) == 101

    assert len(ossfs.find(path + "prefixes/test_")) == 0
    assert len(ossfs.find(path + "prefixes", prefix="test_")) == 100
    assert len(ossfs.find(path + "prefixes/", prefix="test_")) == 100

    test_1s = ossfs.find(path + "prefixes/test_1")
    assert len(test_1s) == 1
    assert test_1s[0] == path + "prefixes/test_1"

    test_1s = ossfs.find(path + "prefixes/", prefix="test_1")
    assert len(test_1s) == 11
    assert test_1s == [path + "prefixes/test_1"] + [
        path + f"prefixes/test_{cursor}" for cursor in range(10, 20)
    ]


WRITE_BLOCK_SIZE = 2**14  # 16KB blocks
READ_BLOCK_SIZE = 2**14  # 16KB blocks


@pytest.mark.parametrize("ossfs", ["sync"], indirect=True)
def test_get_put_file(
    ossfs: Union["OSSFileSystem", "AioOSSFileSystem"], tmpdir, test_path: str
):
    path = f"{test_path}/{function_name(ossfs)}/"
    src_file = str(tmpdir / "source")
    src2_file = str(tmpdir / "source_2")
    dest_file = path + "get_put_file/dest"

    data = b"test" * 2**20

    class EventLogger(fsspec.Callback):
        def __init__(self):
            self.events = []
            super().__init__()

        def set_size(self, size):
            self.events.append(("set_size", size))

        def absolute_update(self, value):
            self.events.append(("absolute_update", value))

    with open(src_file, "wb") as f_write:
        f_write.write(data)

    event_logger = EventLogger()
    ossfs.put_file(src_file, dest_file, callback=event_logger)
    assert ossfs.exists(dest_file)

    assert event_logger.events[0] == ("set_size", len(data))
    assert len(event_logger.events[1:]) == len(data) // WRITE_BLOCK_SIZE

    event_logger = EventLogger()
    ossfs.get_file(dest_file, src2_file, callback=event_logger)
    with open(src2_file, "rb") as f_read:
        assert f_read.read() == data

    assert event_logger.events[0] == ("set_size", len(data))
    assert len(event_logger.events[1:]) == len(data) // READ_BLOCK_SIZE



"""
Code of base class of OSSFileSystem
"""

import logging
import os
import re
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Set, Tuple

from fsspec.spec import AbstractFileSystem
from fsspec.utils import stringify_path
from oss2.auth import AnonymousAuth, Auth, StsAuth
from oss2.defaults import multiget_threshold

from .file import OSSFile

logger = logging.getLogger("ossfs")
logging.getLogger("oss2").setLevel(logging.CRITICAL)
logging.getLogger("urllib3").setLevel(logging.WARNING)

if TYPE_CHECKING:
    from oss2.models import SimplifiedObjectInfo


DEFAULT_POOL_SIZE = 20
DEFAULT_BLOCK_SIZE = 5 * 2**20
SIMPLE_TRANSFER_THRESHOLD = multiget_threshold


class BaseOSSFileSystem(AbstractFileSystem):
    """
    base class of the ossfs (ossfs) OSS file system access OSS(Object
    Storage Service) as if it were a file system.

    This exposes a filesystem-like API (ls, cp, open, etc.) on top of OSS
    storage.

    Provide credentials with `key` and `secret`, or together with `token`.
    If anonymous leave all these argument empty.

    """

    protocol = "oss"

    def __init__(  # noqa: PLR0913
        self,
        endpoint: Optional[str] = None,
        key: Optional[str] = None,
        secret: Optional[str] = None,
        token: Optional[str] = None,
        default_cache_type: str = "readahead",
        default_block_size: Optional[int] = None,
        **kwargs,
    ):
        """
        Parameters
        ----------
        endpoint: string (None)
            Default endpoints of the fs Endpoints are the address where OSS
            locate like: http://oss-cn-hangzhou.aliyuncs.com or
            https://oss-me-east-1.aliyuncs.com, Can be changed after the
            initialization.
        key : string (None)
            If not anonymous, use this access key ID, if specified.
        secret : string (None)
            If not anonymous, use this secret access key, if specified.
        token : string (None)
            If not anonymous, use this security token, if specified.
        default_block_size: int (None)
            If given, the default block size value used for ``open()``, if no
            specific value is given at all time. The built-in default is 5MB.
        default_cache_type : string ("readahead")
            If given, the default cache_type value used for ``open()``. Set to "none"
            if no caching is desired. See fsspec's documentation for other available
            cache_type values. Default cache_type is "readahead".

        The following parameters are passed on to fsspec:

        skip_instance_cache: to control reuse of instances
        use_listings_cache, listings_expiry_time, max_paths: to control reuse of
        directory listings
        """
        if token:
            self._auth = StsAuth(key, secret, token)
        elif key:
            self._auth = Auth(key, secret)
        else:
            self._auth = AnonymousAuth()
        self._endpoint = endpoint or os.getenv("OSS_ENDPOINT")
        if self._endpoint is None:
            logger.warning(
                "OSS endpoint is not set, OSSFS could not work properly"
                "without a endpoint, please set it manually with "
                "`ossfs.set_endpoint` later"
            )

        super_kwargs = {
            k: kwargs.pop(k)
            for k in ["use_listings_cache", "listings_expiry_time", "max_paths"]
            if k in kwargs
        }  # passed to fsspec superclass
        super().__init__(**super_kwargs)

        self._default_block_size = default_block_size or DEFAULT_BLOCK_SIZE
        self._default_cache_type = default_cache_type

    def set_endpoint(self, endpoint: str):
        """
        Reset the endpoint for ossfs
        endpoint : string (None)
            Default endpoints of the fs
            Endpoints are the address where OSS locate
            like: http://oss-cn-hangzhou.aliyuncs.com or
        """
        if not endpoint:
            raise ValueError("Not a valid endpoint")
        self._endpoint = endpoint

    @classmethod
    def _strip_protocol(cls, path):
        """Turn path from fully-qualified to file-system-specifi
        Parameters
        ----------
        path : Union[str, List[str]]
            Input path, like
            `http://oss-cn-hangzhou.aliyuncs.com/mybucket/myobject`
            `oss://mybucket/myobject`
        Examples
        --------
        >>> _strip_protocol(
            "http://oss-cn-hangzhou.aliyuncs.com/mybucket/myobject"
            )
        ('/mybucket/myobject')
        >>> _strip_protocol(
            "oss://mybucket/myobject"
            )
        ('/mybucket/myobject')
        """
        if isinstance(path, list):
            return [cls._strip_protocol(p) for p in path]
        path_string: str = stringify_path(path)
        if path_string.startswith("oss://"):
            path_string = path_string[5:]

        parser_re = r"https?://(?P<endpoint>oss.+aliyuncs\.com)(?P<path>/.+)"
        matcher = re.compile(parser_re).match(path_string)
        if matcher:
            path_string = matcher["path"]
        return path_string or cls.root_marker

    def split_path(self, path: str) -> Tuple[str, str]:
        """
        Normalise object path string into bucket and key.
        Parameters
        ----------
        path : string
            Input path, like `/mybucket/path/to/file`
        Examples
        --------
        >>> split_path("/mybucket/path/to/file")
        ['mybucket', 'path/to/file' ]
        >>> split_path("
        /mybucket/path/to/versioned_file?versionId=some_version_id
        ")
        ['mybucket', 'path/to/versioned_file', 'some_version_id']
        """
        path = self._strip_protocol(path)
        path = path.lstrip("/")
        if "/" not in path:
            return path, ""
        bucket_name, obj_name = path.split("/", 1)
        return bucket_name, obj_name

    def invalidate_cache(self, path: Optional[str] = None):
        if path is None:
            self.dircache.clear()
        else:
            norm_path: str = self._strip_protocol(path)
            norm_path = norm_path.lstrip("/")
            self.dircache.pop(norm_path, None)
            while norm_path:
                self.dircache.pop(norm_path, None)
                norm_path = self._parent(norm_path)

    def _transfer_object_info_to_dict(
        self, bucket: str, obj: "SimplifiedObjectInfo"
    ) -> Dict:
        data: Dict[str, Any] = {
            "name": f"{bucket}/{obj.key}",
            "type": "file",
            "size": obj.size,
        }
        if obj.last_modified:
            data["LastModified"] = obj.last_modified
        if obj.is_prefix():
            data["type"] = "directory"
            data["size"] = 0
        return data

    def _verify_find_arguments(
        self, path: str, maxdepth: Optional[int], withdirs: bool, prefix: str
    ) -> str:
        path = self._strip_protocol(path)
        bucket, _ = self.split_path(path)
        if not bucket:
            raise ValueError("Cannot traverse all of the buckets")
        if (withdirs or maxdepth) and prefix:
            raise ValueError(
                "Can not specify 'prefix' option alongside "
                "'withdirs'/'maxdepth' options."
            )
        return path

    def _get_batch_delete_key_list(self, pathlist: List[str]) -> Tuple[str, List[str]]:
        buckets: Set[str] = set()
        key_list: List[str] = []
        for path in pathlist:
            bucket, key = self.split_path(path)
            buckets.add(bucket)
            key_list.append(key)
        if len(buckets) > 1:
            raise ValueError("Bulk delete files should refer to only one bucket")
        bucket = buckets.pop()
        if len(pathlist) > 1000:
            raise ValueError("Max number of files to delete in one call is 1000")
        for path in pathlist:
            self.invalidate_cache(self._parent(path))
        return bucket, key_list

    def _open(
        self,
        path: str,
        mode: str = "rb",
        block_size: Optional[int] = None,
        autocommit: bool = True,
        cache_options: Optional[str] = None,
        **kwargs,
    ) -> "OSSFile":
        """
        Open a file for reading or writing.
        Parameters
        ----------
        path: str
            File location
        mode: str
            'rb', 'wb', etc.
        autocommit: bool
            If False, writes to temporary file that only gets put in final
            location upon commit
        kwargs
        Returns
        -------
        OSSFile instance
        """
        cache_type = kwargs.pop("cache_type", self._default_cache_type)
        return OSSFile(
            self,
            path,
            mode,
            block_size,
            autocommit,
            cache_options=cache_options,
            cache_type=cache_type,
            **kwargs,
        )

    def touch(self, path: str, truncate: bool = True, **kwargs):
        """Create empty file, or update timestamp

        Parameters
        ----------
        path: str
            file location
        truncate: bool
            If True, always set file size to 0; if False, update timestamp and
            leave file unchanged, if backend allows this
        """
        if truncate or not self.exists(path):
            self.invalidate_cache(self._parent(path))
            with self.open(path, "wb", **kwargs):
                pass
        else:
            raise NotImplementedError



"""
Test benchmark of get and push
"""

import os
from typing import TYPE_CHECKING, Union

import pytest

from tests.conftest import bucket_relative_path, function_name

if TYPE_CHECKING:
    from oss2 import Bucket

    from ossfs import AioOSSFileSystem, OSSFileSystem


@pytest.mark.parametrize("ossfs", ["sync", "async"], indirect=True)
def test_put_small_files(
    ossfs: Union["OSSFileSystem", "AioOSSFileSystem"],
    tmpdir,
    test_path: str,
    benchmark,
):
    path = f"{test_path}/{function_name(ossfs)}/"

    local_path = tmpdir / "number"
    local_path.mkdir()

    for num in range(1200):
        data = os.urandom(1000)
        local_file = local_path / str(num)
        with open(local_file, "wb") as f_w:
            f_w.write(data)

    benchmark.pedantic(
        ossfs.put,
        args=(str(local_path) + "/*", path),
        kwargs={},
        iterations=1,
        rounds=1,
    )
    assert len(ossfs.ls(path + "number/")) == 1200


@pytest.mark.parametrize("ossfs", ["sync", "async"], indirect=True)
def test_get_small_files(
    ossfs: Union["OSSFileSystem", "AioOSSFileSystem"],
    tmpdir,
    bucket: "Bucket",
    test_path: str,
    benchmark,
):
    path = f"{test_path}/{function_name(ossfs)}/"

    for num in range(1200):
        data = os.urandom(1000)
        key = bucket_relative_path(path + str(num))
        bucket.put_object(key, data)

    local_path = tmpdir / "number"
    local_path.mkdir()
    local_dir = local_path / ossfs.__class__.__name__
    local_dir.mkdir()

    benchmark.pedantic(
        ossfs.get, args=(path + "*", str(local_path)), kwargs={}, iterations=1, rounds=1
    )
    assert len(local_dir.listdir()) == 1200



"""Automation using nox."""

import glob
import os

import nox

nox.options.reuse_existing_virtualenvs = True
nox.options.sessions = "lint", "tests"
locations = "src", "tests"


@nox.session(
    python=["3.8", "3.9", "3.10", "3.11", "3.12", "3.13", "pypy3.8", "pypy3.9"]
)
def tests(session: nox.Session) -> None:
    session.install(".[tests]")
    session.run(
        "pytest",
        "tests/func",
        "--cov",
        "--cov-config=pyproject.toml",
        "--durations=100",
        *session.posargs,
        env={"COVERAGE_FILE": f".coverage.{session.python}"},
    )


@nox.session
def bench(session: nox.Session) -> None:
    session.install(".[tests]")
    session.run("pytest", "--benchmark-only", "tests/bench")


@nox.session
def lint(session: nox.Session) -> None:
    session.install("pre-commit")
    session.install("-e", ".[dev]")

    args = *(session.posargs or ("--show-diff-on-failure",)), "--all-files"
    session.run("pre-commit", "run", *args)
    session.run("python", "-m", "mypy")


@nox.session
def build(session: nox.Session) -> None:
    session.install("build", "twine")
    session.run("python", "-m", "build")
    dists = glob.glob("dist/*")
    session.run("twine", "check", *dists, silent=True)


@nox.session
def dev(session: nox.Session) -> None:
    """Sets up a python development environment for the project."""
    args = session.posargs or ("venv",)
    venv_dir = os.fsdecode(os.path.abspath(args[0]))

    session.log(f"Setting up virtual environment in {venv_dir}")
    session.install("virtualenv")
    session.run("virtualenv", venv_dir, silent=True)

    python = os.path.join(venv_dir, "bin/python")
    session.run(python, "-m", "pip", "install", "-e", ".[dev]", external=True)



"""utils of ossfs"""

import copy
import inspect
from functools import wraps
from typing import Any, Dict, List


def _copy_and_pretify_list(
    path: str, result: List[Dict[str, Any]], detail: bool
) -> List[Dict[str, Any]]:
    result_copy = copy.deepcopy(result)
    if not detail:
        for path_info in result_copy:
            if path.startswith("/"):
                path_info["name"] = f"/{path_info['name'].lstrip('/')}"
            else:
                path_info["name"] = f"{path_info['name'].lstrip('/')}"
        final_results = sorted(info["name"] for info in result_copy)
    else:
        for path_info in result_copy:
            path_info["Size"] = path_info["size"]
            path_info["Key"] = path_info["name"]
            if path.startswith("/"):
                path_info["name"] = f"/{path_info['name'].lstrip('/')}"
                path_info["Key"] = f"/{path_info['Key'].lstrip('/')}"
            else:
                path_info["name"] = f"{path_info['name'].lstrip('/')}"
                path_info["Key"] = f"{path_info['Key'].lstrip('/')}"
        final_results = sorted(result_copy, key=lambda i: i["name"])
    return final_results


def _format_unify(path: str, result, detail: bool):
    if not result:
        return result
    if isinstance(result, dict):
        for value in result.values():
            nested = isinstance(value, dict)
            break
        if nested:
            result_list = [path_info for _, path_info in result.items()]
            normed_result = _copy_and_pretify_list(path, result_list, detail)
            if detail:
                return {path_info["name"]: path_info for path_info in normed_result}
            return normed_result

        return _copy_and_pretify_list(path, [result], detail)[0]

    return _copy_and_pretify_list(path, result, detail)


def prettify_info_result(func):
    """Make the return values of `ls` and `info` follows the fsspec's standard
    Examples:
    --------------------------------
    @pretify_info_result
    def ls(path: str, ...)
    """

    @wraps(func)
    def wrapper(ossfs, path: str, *args, **kwargs):
        func_params = inspect.signature(func).parameters
        if "detail" in func_params:
            detail = kwargs.get("detail", func_params["detail"].default)
        else:
            detail = kwargs.get("detail", True)

        result = func(ossfs, path, *args, **kwargs)
        return _format_unify(path, result, detail)

    return wrapper


def async_prettify_info_result(func):
    """Make the return values of async func `ls` and `info` follows the
    fsspec's standard
    Examples:
    --------------------------------
    @async_pretify_info_result
    async def ls(path: str, ...)
    """

    @wraps(func)
    async def wrapper(ossfs, path: str, *args, **kwargs):
        func_params = inspect.signature(func).parameters
        if "detail" in func_params:
            detail = kwargs.get("detail", func_params["detail"].default)
        else:
            detail = kwargs.get("detail", True)
        result = await func(ossfs, path, *args, **kwargs)
        return _format_unify(path, result, detail)

    return wrapper


def as_progress_handler(callback):
    """progress bar handler"""
    if callback is None:
        return None

    sent_total = False

    def progress_handler(absolute_progress, total_size):
        nonlocal sent_total
        if not sent_total:
            callback.set_size(total_size)
            sent_total = True

        callback.absolute_update(absolute_progress)

    return progress_handler
